# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test mongo using the synchronizer, i.e. as it would be used by an
    user
"""

import os
import sys
import time

from bson import SON
from gridfs import GridFS

sys.path[0:0] = [""]  # noqa

from mongo_connector.doc_managers.mongo_doc_manager import DocManager
from mongo_connector.connector import Connector
from mongo_connector.util import retry_until_ok
from mongo_connector.test_utils import (
    ReplicaSet,
    ReplicaSetSingle,
    Server,
    connector_opts,
    assert_soon,
    close_client,
)

from tests import unittest, SkipTest
from mongo_connector.version import Version


class MongoTestCase(unittest.TestCase):

    use_single_meta_collection = False

    @classmethod
    def setUpClass(cls):
        cls.standalone = Server().start()
        cls.mongo_doc = DocManager(cls.standalone.uri)
        cls.mongo_conn = cls.standalone.client()
        cls.desination_version = Version.from_client(cls.mongo_conn)
        cls.mongo = cls.mongo_conn["test"]["test"]

    @classmethod
    def tearDownClass(cls):
        close_client(cls.mongo_conn)
        cls.standalone.stop()

    def _search(self, **kwargs):
        for doc in self.mongo.find(**kwargs):
            yield doc

        fs = GridFS(self.mongo_conn["test"], "test")

        collection_name = "test.test"
        if self.use_single_meta_collection:
            collection_name = "__oplog"
        col = self.mongo_conn["__mongo_connector"][collection_name]
        for doc in col.find():
            if doc.get("gridfs_id"):
                for f in fs.find({"_id": doc["gridfs_id"]}):
                    doc["filename"] = f.filename
                    doc["content"] = f.read()
                    yield doc

    def _remove(self):
        for db in self.mongo_conn.database_names():
            if db not in ["local", "admin"]:
                self.mongo_conn.drop_database(db)


class MongoReplicaSetTestCase(MongoTestCase):
    def setUp(self):
        self.repl_set = self.replica_set_class().start()
        self.conn = self.repl_set.client()

        try:
            os.unlink("oplog.timestamp")
        except OSError:
            pass
        self._remove()
        self.connector = Connector(
            mongo_address=self.repl_set.uri,
            doc_managers=(self.mongo_doc,),
            namespace_options={
                "test.test": {"gridfs": True},
                "rename.me": "new.target",
                "rename.me2": "new2.target2",
            },
            **connector_opts
        )

        self.connector.start()
        assert_soon(lambda: len(self.connector.shard_set) > 0)
        assert_soon(lambda: sum(1 for _ in self._search()) == 0)

    def drop_all_databases(self):
        for name in self.mongo_conn.database_names():
            if name not in ["local", "admin"]:
                self.mongo_conn.drop_database(name)
        for name in self.conn.database_names():
            if name not in ["local", "admin"]:
                self.conn.drop_database(name)

    def tearDown(self):
        self.connector.join()
        self.drop_all_databases()
        self.repl_set.stop()


class TestMongoReplicaSetSingle(MongoReplicaSetTestCase):
    """ Tests MongoDB to MongoDB DocManager replication with a 1 node replica
    set.
    """

    replica_set_class = ReplicaSetSingle

    def test_insert(self):
        """Tests insert
        """

        self.conn["test"]["test"].insert_one({"name": "paulie"})
        assert_soon(lambda: sum(1 for _ in self._search()) == 1)
        result_set_1 = self._search()
        self.assertEqual(sum(1 for _ in result_set_1), 1)
        result_set_2 = self.conn["test"]["test"].find_one()
        for item in result_set_1:
            self.assertEqual(item["_id"], result_set_2["_id"])
            self.assertEqual(item["name"], result_set_2["name"])

    def test_remove(self):
        """Tests remove
        """

        self.conn["test"]["test"].insert_one({"name": "paulie"})
        assert_soon(lambda: sum(1 for _ in self._search()) == 1)
        self.conn["test"]["test"].delete_one({"name": "paulie"})
        assert_soon(lambda: sum(1 for _ in self._search()) != 1)
        self.assertEqual(sum(1 for _ in self._search()), 0)

    def test_insert_file(self):
        """Tests inserting a gridfs file
        """
        fs = GridFS(self.conn["test"], "test")
        test_data = b"test_insert_file test file"
        id = fs.put(test_data, filename="test.txt", encoding="utf8")
        assert_soon(lambda: sum(1 for _ in self._search()) > 0)

        res = list(self._search())
        self.assertEqual(len(res), 1)
        doc = res[0]
        self.assertEqual(doc["filename"], "test.txt")
        self.assertEqual(doc["_id"], id)
        self.assertEqual(doc["content"], test_data)

    def test_remove_file(self):
        fs = GridFS(self.conn["test"], "test")
        id = fs.put("test file", filename="test.txt", encoding="utf8")
        assert_soon(lambda: sum(1 for _ in self._search()) == 1)
        fs.delete(id)
        assert_soon(lambda: sum(1 for _ in self._search()) == 0)

    def test_update(self):
        """Test update operations."""
        # Insert
        self.conn.test.test.insert_one({"a": 0})
        assert_soon(lambda: sum(1 for _ in self._search()) == 1)

        def check_update(update_spec):
            updated = self.conn.test.command(
                SON(
                    [
                        ("findAndModify", "test"),
                        ("query", {"a": 0}),
                        ("update", update_spec),
                        ("new", True),
                    ]
                )
            )["value"]

            def update_worked():
                replicated = self.mongo_doc.mongo.test.test.find_one({"a": 0})
                return replicated == updated

            # Allow some time for update to propagate
            assert_soon(update_worked)

        # Update by adding a field
        check_update({"$set": {"b": [{"c": 10}, {"d": 11}]}})

        # Update by setting an attribute of a sub-document beyond end of array.
        check_update({"$set": {"b.10.c": 42}})

        # Update by un-setting an array element.
        check_update({"$unset": {"b.10": True}})

        # Update by un-setting a non-existent attribute.
        check_update({"$unset": {"not-present": True}})

        # Update by changing a value within a sub-document (contains array)
        check_update({"$inc": {"b.0.c": 1}})

        # Update by changing the value within an array
        check_update({"$inc": {"b.1.f": 12}})

        # Update by adding new bucket to list
        check_update({"$push": {"b": {"e": 12}}})

        # Update by changing an entire sub-document
        check_update({"$set": {"b.0": {"e": 4}}})

        # Update by adding a sub-document
        check_update({"$set": {"b": {"0": {"c": 100}}}})

        # Update whole document
        check_update({"a": 0, "b": {"1": {"d": 10000}}})

    def check_renamed_insert(self, target_coll):
        target_db, target_coll = target_coll.split(".", 1)
        mongo_target = self.mongo_conn[target_db][target_coll]
        assert_soon(lambda: len(list(mongo_target.find({}))))
        target_docs = list(mongo_target.find({}))
        self.assertEqual(len(target_docs), 1)
        self.assertEqual(target_docs[0]["renamed"], 1)

    def create_renamed_collection(self, source_coll, target_coll):
        """Create renamed collections for the command tests."""
        # Create the rename database and 'rename.me' collection
        source_db, source_coll = source_coll.split(".", 1)
        mongo_source = self.conn[source_db][source_coll]
        mongo_source.insert_one({"renamed": 1})
        self.check_renamed_insert(target_coll)

    def test_drop_database_renamed(self):
        """Test the dropDatabase command on a renamed database."""
        if not self.desination_version.at_least(3, 0, 7) or (
            self.desination_version.at_least(3, 1)
            and not self.desination_version.at_least(3, 1, 9)
        ):
            raise SkipTest("This test fails often because of SERVER-13212")
        self.create_renamed_collection("rename.me", "new.target")
        self.create_renamed_collection("rename.me2", "new2.target2")
        # test that drop database removes target databases
        self.conn.drop_database("rename")
        assert_soon(lambda: "new" not in self.mongo_conn.database_names())
        assert_soon(lambda: "new2" not in self.mongo_conn.database_names())

    def test_drop_collection_renamed(self):
        """Test the drop collection command on a renamed collection."""
        self.create_renamed_collection("rename.me", "new.target")
        self.create_renamed_collection("rename.me2", "new2.target2")
        # test that drop collection removes target collection
        self.conn.rename.drop_collection("me")
        assert_soon(lambda: "target" not in self.mongo_conn.new.collection_names())
        self.conn.rename.drop_collection("me2")
        assert_soon(lambda: "target2" not in self.mongo_conn.new2.collection_names())

    def test_rename_collection_renamed(self):
        """Test the renameCollection command on a renamed collection to a
        renamed collection.
        """
        self.create_renamed_collection("rename.me", "new.target")
        self.conn.admin.command("renameCollection", "rename.me", to="rename.me2")
        # In the target, 'new.target' should be renamed to 'new2.target2'
        assert_soon(lambda: "target" not in self.mongo_conn.new.collection_names())
        self.check_renamed_insert("new2.target2")


class TestMongoReplicaSet(MongoReplicaSetTestCase):
    """ Tests MongoDB to MongoDB DocManager replication with a 3 node replica
    set.
    """

    replica_set_class = ReplicaSet

    def test_rollback(self):
        """Tests rollback. We force a rollback by adding a doc, killing the
        primary, adding another doc, killing the new primary, and then
        restarting both.
        """
        primary_conn = self.repl_set.primary.client()
        self.conn["test"]["test"].insert_one({"name": "paul"})
        condition = (
            lambda: self.conn["test"]["test"].find_one({"name": "paul"}) is not None
        )
        assert_soon(condition)
        assert_soon(lambda: sum(1 for _ in self._search()) == 1)

        self.repl_set.primary.stop(destroy=False)
        new_primary_conn = self.repl_set.secondary.client()
        admin = new_primary_conn["admin"]

        def condition():
            return admin.command("isMaster")["ismaster"]

        assert_soon(lambda: retry_until_ok(condition))

        retry_until_ok(self.conn.test.test.insert_one, {"name": "pauline"})
        assert_soon(lambda: sum(1 for _ in self._search()) == 2)
        result_set_1 = list(self._search())
        result_set_2 = self.conn["test"]["test"].find_one({"name": "pauline"})
        self.assertEqual(len(result_set_1), 2)
        # make sure pauline is there
        for item in result_set_1:
            if item["name"] == "pauline":
                self.assertEqual(item["_id"], result_set_2["_id"])
        self.repl_set.secondary.stop(destroy=False)

        self.repl_set.primary.start()
        assert_soon(lambda: primary_conn["admin"].command("isMaster")["ismaster"])

        self.repl_set.secondary.start()

        time.sleep(2)
        result_set_1 = list(self._search())
        self.assertEqual(len(result_set_1), 1)
        for item in result_set_1:
            self.assertEqual(item["name"], "paul")
        find_cursor = retry_until_ok(self.conn["test"]["test"].find)
        self.assertEqual(retry_until_ok(find_cursor.count), 1)


if __name__ == "__main__":
    unittest.main()



import base64
import datetime
import re

from uuid import UUID
from math import isnan, isinf

import logging

import bson
import bson.json_util

long = int
unicode = str

LOG = logging.getLogger(__name__)

RE_TYPE = type(re.compile(""))
try:
    from bson.regex import Regex

    RE_TYPES = (RE_TYPE, Regex)
except ImportError:
    RE_TYPES = (RE_TYPE,)


class DocumentFormatter(object):
    """Interface for classes that can transform documents to conform to
    external drivers' expectations.
    """

    def transform_value(self, value):
        """Transform a leaf-node in a document.

        This method may be overridden to provide custom handling for specific
        types of values.
        """
        raise NotImplementedError

    def transform_element(self, key, value):
        """Transform a single key-value pair within a document.

        This method may be overridden to provide custom handling for specific
        types of values. This method should return an iterator over the
        resulting key-value pairs.
        """
        raise NotImplementedError

    def format_document(self, document):
        """Format a document in preparation to be sent to an external driver."""
        raise NotImplementedError


class DefaultDocumentFormatter(DocumentFormatter):
    """Basic DocumentFormatter that preserves numbers, base64-encodes binary,
    and stringifies everything else.
    """

    def transform_value(self, value):
        # This is largely taken from bson.json_util.default, though not the same
        # so we don't modify the structure of the document
        if isinstance(value, dict):
            return self.format_document(value)
        elif isinstance(value, list):
            return [self.transform_value(v) for v in value]
        if isinstance(value, RE_TYPES):
            flags = ""
            if value.flags & re.IGNORECASE:
                flags += "i"
            if value.flags & re.LOCALE:
                flags += "l"
            if value.flags & re.MULTILINE:
                flags += "m"
            if value.flags & re.DOTALL:
                flags += "s"
            if value.flags & re.UNICODE:
                flags += "u"
            if value.flags & re.VERBOSE:
                flags += "x"
            pattern = value.pattern
            # quasi-JavaScript notation (may include non-standard flags)
            return "/%s/%s" % (pattern, flags)
        elif isinstance(value, bson.Binary) or isinstance(value, bytes):
            # Just include body of binary data without subtype
            return base64.b64encode(value).decode()
        elif isinstance(value, UUID):
            return value.hex
        elif isinstance(value, (int, long, float)):
            if isnan(value):
                raise ValueError("nan")
            elif isinf(value):
                raise ValueError("inf")
            return value
        elif isinstance(value, datetime.datetime):
            return value
        elif value is None:
            return value
        # Default
        return unicode(value)

    def transform_element(self, key, value):
        try:
            new_value = self.transform_value(value)
            yield key, new_value
        except ValueError as e:
            LOG.warn("Invalid value for key: %s as %s" % (key, str(e)))

    def format_document(self, document):
        def _kernel(doc):
            for key in doc:
                value = doc[key]
                for new_k, new_v in self.transform_element(key, value):
                    yield new_k, new_v

        return dict(_kernel(document))


class DocumentFlattener(DefaultDocumentFormatter):
    """Formatter that completely flattens documents and unwinds arrays:

    An example:
      {"a": 2,
       "b": {
         "c": {
           "d": 5
         }
       },
       "e": [6, 7, 8]
      }

    becomes:
      {"a": 2, "b.c.d": 5, "e.0": 6, "e.1": 7, "e.2": 8}

    """

    def transform_element(self, key, value):
        if isinstance(value, list):
            for li, lv in enumerate(value):
                for inner_k, inner_v in self.transform_element("%s.%s" % (key, li), lv):
                    yield inner_k, inner_v
        elif isinstance(value, dict):
            formatted = self.format_document(value)
            for doc_key in formatted:
                yield "%s.%s" % (key, doc_key), formatted[doc_key]
        else:
            # We assume that transform_value will return a 'flat' value,
            # not a list or dict
            yield key, self.transform_value(value)

    def format_document(self, document):
        def flatten(doc, path):
            top_level = len(path) == 0
            if not top_level:
                path_string = ".".join(path)
            for k in doc:
                v = doc[k]
                if isinstance(v, dict):
                    path.append(k)
                    for inner_k, inner_v in flatten(v, path):
                        yield inner_k, inner_v
                    path.pop()
                else:
                    transformed = self.transform_element(k, v)
                    for new_k, new_v in transformed:
                        if top_level:
                            yield new_k, new_v
                        else:
                            yield "%s.%s" % (path_string, new_k), new_v

        return dict(flatten(document, []))



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Preprocesses the oplog command entries.
"""

import logging

from mongo_connector.namespace_config import NamespaceConfig

LOG = logging.getLogger(__name__)


class CommandHelper(object):
    def __init__(self, namespace_config=None):
        if namespace_config is None:
            namespace_config = NamespaceConfig()
        self.namespace_config = namespace_config

    # Applies the namespace mapping to a database.
    # Individual collections in a database can be mapped to
    # different target databases, so map_db can return multiple results.
    # The input parameter db is plain text
    def map_db(self, db):
        return self.namespace_config.map_db(db)

    # Applies the namespace mapping to a "db.collection" string
    # The input parameter ns is plain text
    def map_namespace(self, ns):
        return self.namespace_config.map_namespace(ns)

    # Applies the namespace mapping to a db and collection
    # The input parameter db and coll are plain text
    def map_collection(self, db, coll):
        ns = self.map_namespace(db + "." + coll)
        if ns:
            return tuple(ns.split(".", 1))
        else:
            return None, None



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests methods for mongo_connector
"""

import json
import os
import sys

from bson.timestamp import Timestamp

sys.path[0:0] = [""]  # noqa

from mongo_connector.connector import Connector, get_mininum_mongodb_version
from mongo_connector.test_utils import (
    ReplicaSetSingle,
    connector_opts,
    assert_soon,
    db_user,
    db_password,
)
from mongo_connector.util import long_to_bson_ts
from mongo_connector.version import Version

from tests import unittest, SkipTest


class TestMongoConnector(unittest.TestCase):
    """ Test Class for the Mongo Connector
    """

    @classmethod
    def setUpClass(cls):
        """ Initializes the cluster
        """
        try:
            os.unlink("oplog.timestamp")
        except OSError:
            pass
        open("oplog.timestamp", "w").close()
        cls.repl_set = ReplicaSetSingle().start()

    @classmethod
    def tearDownClass(cls):
        """ Kills cluster instance
        """
        cls.repl_set.stop()

    def test_connector(self):
        """Test whether the connector initiates properly
        """
        conn = Connector(mongo_address=self.repl_set.uri, **connector_opts)
        conn.start()
        assert_soon(lambda: bool(conn.shard_set))

        # Make sure get_mininum_mongodb_version returns the current version.
        self.assertEqual(
            Version.from_client(self.repl_set.client()), get_mininum_mongodb_version()
        )

        conn.join()

        # Make sure the connector is shutdown correctly
        self.assertFalse(conn.can_run)
        for thread in conn.shard_set.values():
            self.assertFalse(thread.running)

    def test_copy_uri_options(self):
        """Test copy_uri_options returns proper MongoDB URIs."""
        uri = "mongodb://host:27017/db?maxPoolSize=1234&w=2"
        self.assertEqual(
            Connector.copy_uri_options("a:123,[::1]:321", uri),
            "mongodb://a:123,[::1]:321/?maxPoolSize=1234&w=2",
        )
        uri = "host:27017"
        self.assertEqual(
            Connector.copy_uri_options("a:123,[::1]:321", uri),
            "mongodb://a:123,[::1]:321",
        )

    def test_write_oplog_progress(self):
        """Test write_oplog_progress under several circumstances
        """
        try:
            os.unlink("temp_oplog.timestamp")
        except OSError:
            pass
        open("temp_oplog.timestamp", "w").close()
        conn = Connector(
            mongo_address=self.repl_set.uri,
            oplog_checkpoint="temp_oplog.timestamp",
            **connector_opts
        )

        # test that None is returned if there is no config file specified.
        self.assertEqual(conn.write_oplog_progress(), None)

        conn.oplog_progress.get_dict()[1] = Timestamp(12, 34)
        # pretend to insert a thread/timestamp pair
        conn.write_oplog_progress()

        data = json.load(open("temp_oplog.timestamp", "r"))
        self.assertEqual(1, int(data[0]))
        self.assertEqual(long_to_bson_ts(int(data[1])), Timestamp(12, 34))

        # ensure the temp file was deleted
        self.assertFalse(os.path.exists("temp_oplog.timestamp" + "~"))

        # ensure that updates work properly
        conn.oplog_progress.get_dict()[1] = Timestamp(44, 22)
        conn.write_oplog_progress()

        config_file = open("temp_oplog.timestamp", "r")
        data = json.load(config_file)
        self.assertEqual(1, int(data[0]))
        self.assertEqual(long_to_bson_ts(int(data[1])), Timestamp(44, 22))

        config_file.close()
        os.unlink("temp_oplog.timestamp")

    def test_read_oplog_progress(self):
        """Test read_oplog_progress
        """

        conn = Connector(
            mongo_address=self.repl_set.uri, oplog_checkpoint=None, **connector_opts
        )

        # testing with no file
        self.assertEqual(conn.read_oplog_progress(), None)

        try:
            os.unlink("temp_oplog.timestamp")
        except OSError:
            pass
        open("temp_oplog.timestamp", "w").close()

        conn.oplog_checkpoint = "temp_oplog.timestamp"

        # testing with empty file
        self.assertEqual(conn.read_oplog_progress(), None)

        oplog_dict = conn.oplog_progress.get_dict()

        # add a value to the file, delete the dict, and then read in the value
        oplog_dict["oplog1"] = Timestamp(12, 34)
        conn.write_oplog_progress()
        del oplog_dict["oplog1"]

        self.assertEqual(len(oplog_dict), 0)

        conn.read_oplog_progress()
        oplog_dict = conn.oplog_progress.get_dict()

        self.assertTrue("oplog1" in oplog_dict.keys())
        self.assertTrue(oplog_dict["oplog1"], Timestamp(12, 34))

        oplog_dict["oplog1"] = Timestamp(55, 11)

        # see if oplog progress dict is properly updated
        conn.read_oplog_progress()
        self.assertTrue(oplog_dict["oplog1"], Timestamp(55, 11))

        os.unlink("temp_oplog.timestamp")

    def test_connector_minimum_privileges(self):
        """Test the Connector works with a user with minimum privileges."""
        if not (db_user and db_password):
            raise SkipTest("Need to set a user/password to test this.")
        client = self.repl_set.client()
        minimum_user = "read_local_and_included_databases"
        minimum_pwd = "password"
        client.admin.add_user(
            minimum_user,
            minimum_pwd,
            roles=[
                {"role": "read", "db": "test"},
                {"role": "read", "db": "wildcard"},
                {"role": "read", "db": "local"},
            ],
        )

        client.test.test.insert_one({"replicated": 1})
        client.test.ignored.insert_one({"replicated": 0})
        client.ignored.ignored.insert_one({"replicated": 0})
        client.wildcard.test.insert_one({"replicated": 1})
        conn = Connector(
            mongo_address=self.repl_set.primary.uri,
            auth_username=minimum_user,
            auth_key=minimum_pwd,
            namespace_options={"test.test": True, "wildcard.*": True},
        )
        conn.start()
        try:
            assert_soon(conn.doc_managers[0]._search)
        finally:
            conn.join()


if __name__ == "__main__":
    unittest.main()



# Copyright 2013-2016 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import re
import sys
import uuid

import bson

sys.path[0:0] = [""]  # noqa

from mongo_connector.doc_managers.formatters import (
    DefaultDocumentFormatter,
    DocumentFlattener,
)
from tests import unittest


class TestFormatters(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Some test values to use
        cls.bin1 = bson.Binary(b"\x00hello\x00", 0)
        cls.bin2 = b"\x00hello\x00"
        cls.xuuid = uuid.uuid4()
        cls.oid = bson.ObjectId()
        cls.regex = re.compile("hello", re.VERBOSE | re.MULTILINE)
        cls.lst = [cls.regex, cls.bin1, cls.bin2, cls.xuuid, cls.oid]
        cls.date = datetime.datetime.now()
        cls.doc = {
            "r": cls.regex,
            "b1": cls.bin1,
            "b2": cls.bin2,
            "uuid": cls.xuuid,
            "oid": cls.oid,
            "d": cls.date,
        }
        cls.doc_nested = {"doc": cls.doc}
        cls.doc_list = {"list": [cls.doc, cls.doc_nested, cls.lst]}

    def test_types(self):
        trans = DefaultDocumentFormatter().transform_value

        # regex
        _, patt, flags = trans(self.regex).rsplit("/")
        self.assertIn("x", flags)
        self.assertIn("m", flags)
        self.assertNotIn("l", flags)
        self.assertEqual(patt, "hello")

        # binary
        self.assertEqual(trans(self.bin1), "AGhlbGxvAA==")
        self.assertEqual(trans(self.bin2), "AGhlbGxvAA==")

        # datetime
        self.assertEqual(trans(self.date), self.date)

        # UUID
        self.assertEqual(trans(self.xuuid), self.xuuid.hex)

        # Other type
        self.assertEqual(trans(self.oid), str(self.oid))

        # Compound types
        transformed = trans(self.doc)
        for k, v in self.doc.items():
            self.assertEqual(trans(v), transformed[k])
        for el1, el2 in zip(self.lst, map(trans, self.lst)):
            self.assertEqual(trans(el1), el2)

        # Infinity/NaN
        self.assertRaises(ValueError, trans, float("inf"))
        self.assertRaises(ValueError, trans, float("nan"))

    def test_default_formatter(self):
        formatter = DefaultDocumentFormatter()

        def check_format(document):
            transformed = dict(
                (k, formatter.transform_value(v)) for k, v in document.items()
            )
            self.assertEqual(transformed, formatter.format_document(document))

        # Flat
        check_format(self.doc)

        # Nested
        check_format(self.doc_nested)

        # With a list
        check_format(self.doc_list)

    def test_flattener(self):
        formatter = DocumentFlattener()

        # Flat already
        transformed = dict(
            (k, formatter.transform_value(v)) for k, v in self.doc.items()
        )
        self.assertEqual(transformed, formatter.format_document(self.doc))

        # Nested
        transformed2 = formatter.format_document(self.doc_nested)
        constructed = dict(
            ("doc.%s" % k, formatter.transform_value(v)) for k, v in self.doc.items()
        )
        self.assertEqual(transformed2, constructed)

        # With a list
        constructed1 = dict(
            ("list.0.%s" % k, formatter.transform_value(v)) for k, v in self.doc.items()
        )
        constructed2 = dict(("list.1.%s" % k, v) for k, v in transformed2.items())
        constructed3 = dict(
            ("list.2.%d" % i, formatter.transform_value(v))
            for i, v in enumerate(self.lst)
        )
        constructed1.update(constructed2)
        constructed1.update(constructed3)
        self.assertEqual(formatter.format_document(self.doc_list), constructed1)


if __name__ == "__main__":
    unittest.main()



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

from mongo_connector.connector import get_mininum_mongodb_version
from mongo_connector.errors import UpdateDoesNotApply


LOG = logging.getLogger(__name__)


class DocManagerBase(object):
    """Base class for all DocManager implementations."""

    def apply_update(self, doc, update_spec):
        """Apply an update operation to a document."""

        # Helper to cast a key for a list or dict, or raise ValueError
        def _convert_or_raise(container, key):
            if isinstance(container, dict):
                return key
            elif isinstance(container, list):
                return int(key)
            else:
                raise ValueError

        # Helper to retrieve (and/or create)
        # a dot-separated path within a document.
        def _retrieve_path(container, path, create=False):
            looking_at = container
            for part in path:
                if isinstance(looking_at, dict):
                    if create and part not in looking_at:
                        looking_at[part] = {}
                    looking_at = looking_at[part]
                elif isinstance(looking_at, list):
                    index = int(part)
                    # Do we need to create additional space in the array?
                    if create and len(looking_at) <= index:
                        # Fill buckets with None up to the index we need.
                        looking_at.extend([None] * (index - len(looking_at)))
                        # Bucket we need gets the empty dictionary.
                        looking_at.append({})
                    looking_at = looking_at[index]
                else:
                    raise ValueError
            return looking_at

        def _set_field(doc, to_set, value):
            if "." in to_set:
                path = to_set.split(".")
                where = _retrieve_path(doc, path[:-1], create=True)
                index = _convert_or_raise(where, path[-1])
                wl = len(where)
                if isinstance(where, list) and index >= wl:
                    where.extend([None] * (index + 1 - wl))
                where[index] = value
            else:
                doc[to_set] = value

        def _unset_field(doc, to_unset):
            try:
                if "." in to_unset:
                    path = to_unset.split(".")
                    where = _retrieve_path(doc, path[:-1])
                    index_or_key = _convert_or_raise(where, path[-1])
                    if isinstance(where, list):
                        # Unset an array element sets it to null.
                        where[index_or_key] = None
                    else:
                        # Unset field removes it entirely.
                        del where[index_or_key]
                else:
                    del doc[to_unset]
            except (KeyError, IndexError, ValueError):
                source_version = get_mininum_mongodb_version()
                if source_version is None or source_version.at_least(2, 6):
                    raise
                # Ignore unset errors since MongoDB 2.4 records invalid
                # $unsets in the oplog.
                LOG.warning(
                    "Could not unset field %r from document %r. "
                    "This may be normal when replicating from "
                    "MongoDB 2.4 or the destination could be out of "
                    "sync." % (to_unset, doc)
                )

        # wholesale document replacement
        if "$set" not in update_spec and "$unset" not in update_spec:
            # update spec contains the new document in its entirety
            return update_spec
        else:
            try:
                # $set
                for to_set in update_spec.get("$set", []):
                    value = update_spec["$set"][to_set]
                    _set_field(doc, to_set, value)

                # $unset
                for to_unset in update_spec.get("$unset", []):
                    _unset_field(doc, to_unset)

            except (KeyError, ValueError, AttributeError, IndexError):
                exc_t, exc_v, exc_tb = sys.exc_info()
                msg = "Cannot apply update %r to %r" % (update_spec, doc)
                raise UpdateDoesNotApply(msg).with_traceback(exc_tb)
            return doc

    def bulk_upsert(self, docs, namespace, timestamp):
        """Upsert each document in a set of documents.

        This method may be overridden to upsert many documents at once.
        """
        for doc in docs:
            self.upsert(doc, namespace, timestamp)

    def update(self, doc, update_spec, namespace, timestamp):
        """Update a document.

        ``update_spec`` is the update operation as provided by an oplog record
        in the "o" field.
        """
        raise NotImplementedError

    def upsert(self, document, namespace, timestamp):
        """(Re-)insert a document."""
        raise NotImplementedError

    def remove(self, document_id, namespace, timestamp):
        """Remove a document.

        ``document_id`` is a dict that provides the id of the document
        to be removed. ``namespace`` and ``timestamp`` provide the database +
        collection name and the timestamp from the corresponding oplog entry.
        """
        raise NotImplementedError

    def insert_file(self, f, namespace, timestamp):
        """Insert a file from GridFS."""
        raise NotImplementedError

    def handle_command(self, command_doc, namespace, timestamp):
        """Handle a MongoDB command."""
        raise NotImplementedError

    def search(self, start_ts, end_ts):
        """Get an iterable of documents that were inserted, updated, or deleted
        between ``start_ts`` and ``end_ts``.
        """
        raise NotImplementedError

    def commit(self):
        """Commit all outstanding writes."""
        raise NotImplementedError

    def get_last_doc(self):
        """Get the document that was modified or deleted most recently."""
        raise NotImplementedError

    def stop(self):
        """Stop all threads started by this DocManager."""
        raise NotImplementedError



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test synchronizer using DocManagerSimulator
"""

import os
import sys
import time

sys.path[0:0] = [""]  # noqa

from mongo_connector.connector import Connector
from mongo_connector.namespace_config import NamespaceConfig
from mongo_connector.test_utils import assert_soon, connector_opts, ReplicaSetSingle
from tests import unittest


class TestSynchronizer(unittest.TestCase):
    """ Tests the synchronizers
    """

    @classmethod
    def setUpClass(cls):
        """ Initializes the cluster
        """
        try:
            os.unlink("oplog.timestamp")
        except OSError:
            pass
        open("oplog.timestamp", "w").close()

        cls.repl_set = ReplicaSetSingle().start()
        cls.conn = cls.repl_set.client()
        cls.connector = Connector(
            mongo_address=cls.repl_set.uri, ns_set=["test.test"], **connector_opts
        )
        cls.synchronizer = cls.connector.doc_managers[0]
        cls.connector.start()
        assert_soon(lambda: len(cls.connector.shard_set) != 0)

    @classmethod
    def tearDownClass(cls):
        """ Tears down connector
        """
        cls.connector.join()
        cls.repl_set.stop()

    def setUp(self):
        """ Clears the db
        """
        self.conn["test"]["test"].delete_many({})
        assert_soon(lambda: len(self.synchronizer._search()) == 0)

    def test_insert(self):
        """Tests insert
        """
        self.conn["test"]["test"].insert_one({"name": "paulie"})
        while len(self.synchronizer._search()) == 0:
            time.sleep(1)
        result_set_1 = self.synchronizer._search()
        self.assertEqual(len(result_set_1), 1)
        result_set_2 = self.conn["test"]["test"].find_one()
        for item in result_set_1:
            self.assertEqual(item["_id"], result_set_2["_id"])
            self.assertEqual(item["name"], result_set_2["name"])

    def test_ns_set(self):
        self.conn.test.other.insert_one({"replicated": False})
        results = self.synchronizer._search()
        self.assertEqual(
            len(results), 0, "Should not replicate outside of test.test namespace"
        )

    def test_remove(self):
        """Tests remove
        """
        self.conn["test"]["test"].insert_one({"name": "paulie"})
        while len(self.synchronizer._search()) != 1:
            time.sleep(1)
        self.conn["test"]["test"].delete_one({"name": "paulie"})

        while len(self.synchronizer._search()) == 1:
            time.sleep(1)
        result_set_1 = self.synchronizer._search()
        self.assertEqual(len(result_set_1), 0)

    def test_update(self):
        """Test that Connector can replicate updates successfully."""
        doc = {"a": 1, "b": 2}
        self.conn.test.test.insert_one(doc)
        selector = {"_id": doc["_id"]}

        def update_and_retrieve(update_spec, replace=False):
            if replace:
                self.conn.test.test.replace_one(selector, update_spec)
            else:
                self.conn.test.test.update_one(selector, update_spec)

            # self.conn.test.test.update(selector, update_spec)
            # Give the connector some time to perform update
            time.sleep(1)
            return self.synchronizer._search()[0]

        # Update whole document
        doc = update_and_retrieve({"a": 1, "b": 2, "c": 10}, replace=True)
        self.assertEqual(doc["a"], 1)
        self.assertEqual(doc["b"], 2)
        self.assertEqual(doc["c"], 10)

        # $set only
        doc = update_and_retrieve({"$set": {"b": 4}})
        self.assertEqual(doc["a"], 1)
        self.assertEqual(doc["b"], 4)

        # $unset only
        doc = update_and_retrieve({"$unset": {"a": True}})
        self.assertNotIn("a", doc)
        self.assertEqual(doc["b"], 4)

        # mixed $set/$unset
        doc = update_and_retrieve({"$unset": {"b": True}, "$set": {"c": 3}})
        self.assertEqual(doc["c"], 3)
        self.assertNotIn("b", doc)

        # ensure update works when fields are given
        opthread = self.connector.shard_set[0]
        opthread.namespace_config = NamespaceConfig(include_fields=["a", "b", "c"])
        try:
            doc = update_and_retrieve({"$set": {"d": 10}})
            self.assertEqual(self.conn.test.test.find_one(doc["_id"])["d"], 10)
            self.assertNotIn("d", doc)
            doc = update_and_retrieve({"$set": {"a": 10}})
            self.assertEqual(doc["a"], 10)
        finally:
            # cleanup
            opthread.fields = None


if __name__ == "__main__":
    unittest.main()



# Copyright 2017 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

sys.path[0:0] = [""]  # noqa

from mongo_connector.connector import (
    get_mininum_mongodb_version,
    update_mininum_mongodb_version,
)
from mongo_connector.doc_managers.doc_manager_base import (
    DocManagerBase,
    UpdateDoesNotApply,
)
from mongo_connector.test_utils import TESTARGS
from mongo_connector.version import Version

from tests import unittest


UPDATE_SUCCESS_TEST_CASES = [
    {
        "description": "Update whole document.",
        "doc": {"a": 1},
        "update_spec": {"a": 2, "b": 3},
        "result": {"a": 2, "b": 3},
    },
    {
        "description": "Update by un-setting an existing field.",
        "doc": {"a": 1, "b": 2},
        "update_spec": {"$unset": {"a": True}},
        "result": {"b": 2},
    },
    {
        "description": "Update by un-setting an existing nested field.",
        "doc": {"a": {"b": 2, "c": 3}},
        "update_spec": {"$unset": {"a.b": True}},
        "result": {"a": {"c": 3}},
    },
    {
        "description": "Update by un-setting an array element.",
        "doc": {"a": 1, "b": [0, 1, 2, 3]},
        "update_spec": {"$unset": {"b.1": True}},
        "result": {"a": 1, "b": [0, None, 2, 3]},
    },
    {
        "description": "Update by adding a field.",
        "doc": {"a": 1},
        "update_spec": {"$set": {"b": [{"c": 1}, {"d": 2}]}},
        "result": {"a": 1, "b": [{"c": 1}, {"d": 2}]},
    },
    {
        "description": "Update by adding a nested field.",
        "doc": {"a": 1},
        "update_spec": {"$set": {"b.c.d": 2}},
        "result": {"a": 1, "b": {"c": {"d": 2}}},
    },
    {
        "description": "Update by adding and removing a fields.",
        "doc": {"a": 1, "c": 3},
        "update_spec": {"$unset": {"a": True, "c": True}, "$set": {"b": 2, "d": 4}},
        "result": {"b": 2, "d": 4},
    },
    {
        "description": "Update by setting an element far beyond the end of" "an array.",
        "doc": {"a": 1, "b": [{"c": 1}]},
        "update_spec": {"$set": {"b.4": {"c": 2}}},
        "result": {"a": 1, "b": [{"c": 1}, None, None, None, {"c": 2}]},
    },
    {
        "description": "Update by setting an element right beyond the end "
        "of an array.",
        "doc": {"a": 1, "b": [{"c": 1}]},
        "update_spec": {"$set": {"b.1": {"c": 2}}},
        "result": {"a": 1, "b": [{"c": 1}, {"c": 2}]},
    },
    {
        "description": "Update by setting an attribute of a sub-document far "
        "beyond the end of an array.",
        "doc": {"a": 1, "b": [{"c": 1}]},
        "update_spec": {"$set": {"b.4.c": 2}},
        "result": {"a": 1, "b": [{"c": 1}, None, None, None, {"c": 2}]},
    },
    {
        "description": "Update by setting an attribute of a sub-document "
        "right beyond the end of an array.",
        "doc": {"a": 1, "b": [{"c": 1}]},
        "update_spec": {"$set": {"b.1.c": 2}},
        "result": {"a": 1, "b": [{"c": 1}, {"c": 2}]},
    },
    {
        "description": "Update by changing a field within an array element.",
        "doc": {"a": 1, "b": [{"c": 1}]},
        "update_spec": {"$set": {"b.0.c": 2}},
        "result": {"a": 1, "b": [{"c": 2}]},
    },
    {
        "description": "Update by adding a field within an array element.",
        "doc": {"a": 1, "b": [{"c": 1}, {"d": 2}]},
        "update_spec": {"$set": {"b.1.c": 3}},
        "result": {"a": 1, "b": [{"c": 1}, {"c": 3, "d": 2}]},
    },
    {
        "description": "Update by replacing an array element.",
        "doc": {"a": 1, "b": [0, 1, 2, 3]},
        "update_spec": {"$set": {"b.2": {"new": 2}}},
        "result": {"a": 1, "b": [0, 1, {"new": 2}, 3]},
    },
    {
        "description": "Update by replacing an array with a document with "
        "int field.",
        "doc": {"a": 1, "b": [{"c": 1}, {"d": 2}]},
        "update_spec": {"$set": {"b": {"0": {"e": 100}}}},
        "result": {"a": 1, "b": {"0": {"e": 100}}},
    },
]

UNSET_FAILURE_TEST_CASES = [
    {
        "description": "Update by un-setting a non-existent field.",
        "doc": {"a": 1, "b": 2},
        "update_spec": {"$unset": {"not-present": True}},
        "result": {"a": 1, "b": 2},
    },
    {
        "description": "Update by un-setting a non-existent nested field.",
        "doc": {"a": 1, "b": {"c": {"d": 1}}},
        "update_spec": {"$unset": {"b.not-present.foo": True}},
        "result": {"a": 1, "b": {"c": {"d": 1}}},
    },
    {
        "description": "Update by un-setting invalid array index.",
        "doc": {"a": 1, "b": [0, 1, 2, 3]},
        "update_spec": {"$unset": {"b.not-an-index": True}},
        "result": {"a": 1, "b": [0, 1, 2, 3]},
    },
    {
        "description": "Update by un-setting invalid nested array index.",
        "doc": {"a": 1, "b": [0, 1, 2, 3]},
        "update_spec": {"$unset": {"b.not-an-index.not-present": True}},
        "result": {"a": 1, "b": [0, 1, 2, 3]},
    },
    {
        "description": "Update by un-setting a non-existent array element.",
        "doc": {"a": 1, "b": [0, 1, 2]},
        "update_spec": {"$unset": {"b.4": True}},
        "result": {"a": 1, "b": [0, 1, 2]},
    },
    {
        "description": "Update by un-setting a non-existent field in an array"
        "element.",
        "doc": {"a": 1, "b": [0, {"c": 1}, 2]},
        "update_spec": {"$unset": {"b.1.not-present": True}},
        "result": {"a": 1, "b": [0, {"c": 1}, 2]},
    },
    {
        "description": "Update by adding and removing a non-existent field.",
        "doc": {"a": 1},
        "update_spec": {"$unset": {"a": True, "not-present": True}, "$set": {"b": 2}},
        "result": {"b": 2},
    },
]

UPDATE_FAILURE_TEST_CASES = [
    {
        "description": "Using array notation on non-array field.",
        "doc": {"a": 1},
        "update_spec": {"$set": {"a.0": 2}},
    },
    {
        "description": "Using array notation on non-array field.",
        "doc": {"a": 1},
        "update_spec": {"$set": {"a.0.1": 2}},
    },
    {
        "description": "Using nested field notation on non-object.",
        "doc": {"a": 1},
        "update_spec": {"$set": {"a.b": 2}},
    },
    {
        "description": "Using deeply nested field notation on non-object.",
        "doc": {"a": 1},
        "update_spec": {"$set": {"a.b.c.b": 2}},
    },
    {
        "description": "Setting a field on an array field.",
        "doc": {"a": [{"c": 1}, {"c": 2}]},
        "update_spec": {"$set": {"a.c": 2}},
    },
    {
        "description": "Setting a field on a null array element.",
        "doc": {"a": [None, None]},
        "update_spec": {"$set": {"a.0.c": 2}},
    },
]


class TestDocManagerBase(unittest.TestCase):
    """Unit tests for DocManagerBase"""

    def setUp(self):
        self.base = DocManagerBase()

    def assertUpdateTestSucceeds(self, test):
        self.assertEqual(
            self.base.apply_update(test["doc"], test["update_spec"]),
            test["result"],
            msg=test["description"],
        )

    def assertUpdateTestFails(self, test):
        try:
            doc = self.base.apply_update(test["doc"], test["update_spec"])
            self.fail(
                "UpdateDoesNotApply on MongoDB verison %s not raised for "
                "test: %s, applied %r to %r and got %r"
                % (
                    get_mininum_mongodb_version(),
                    test["description"],
                    test["update_spec"],
                    test["doc"],
                    doc,
                )
            )
        except UpdateDoesNotApply:
            pass

    def test_apply_update(self):
        for test in UPDATE_SUCCESS_TEST_CASES:
            self.assertUpdateTestSucceeds(test)

    def test_apply_update_fails(self):
        for test in UPDATE_FAILURE_TEST_CASES:
            self.assertUpdateTestFails(test)

    def test_apply_update_unset_failures(self):
        # Reset the minimum MongoDB version at the start and end.
        update_mininum_mongodb_version(None)
        for mock_mongodb_version in [(3, 4), (3, 2), (3, 0), (2, 6), (2, 4), None]:
            if mock_mongodb_version is None:
                update_mininum_mongodb_version(None)
            else:
                update_mininum_mongodb_version(Version(*mock_mongodb_version))
            for test in UNSET_FAILURE_TEST_CASES:
                if mock_mongodb_version == (2, 4):
                    self.assertUpdateTestSucceeds(test)
                else:
                    self.assertUpdateTestFails(test)

    def test_bulk_upsert(self):
        with self.assertRaises(NotImplementedError):
            self.base.bulk_upsert([{}], *TESTARGS)

    def test_update(self):
        with self.assertRaises(NotImplementedError):
            self.base.update({}, {}, *TESTARGS)

    def test_upsert(self):
        with self.assertRaises(NotImplementedError):
            self.base.upsert({}, *TESTARGS)

    def test_remove(self):
        with self.assertRaises(NotImplementedError):
            self.base.remove(1, *TESTARGS)

    def test_insert_file(self):
        with self.assertRaises(NotImplementedError):
            self.base.insert_file(None, *TESTARGS)

    def test_handle_command(self):
        with self.assertRaises(NotImplementedError):
            self.base.handle_command({}, *TESTARGS)

    def test_search(self):
        with self.assertRaises(NotImplementedError):
            self.base.search(0, 1)

    def test_commit(self):
        with self.assertRaises(NotImplementedError):
            self.base.commit()

    def test_get_last_doc(self):
        with self.assertRaises(NotImplementedError):
            self.base.get_last_doc()

    def test_stop(self):
        with self.assertRaises(NotImplementedError):
            self.base.stop()


if __name__ == "__main__":
    unittest.main()



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import gridfs

sys.path[0:0] = [""]  # noqa

from mongo_connector import errors
from mongo_connector.gridfs_file import GridFSFile
from mongo_connector.test_utils import ReplicaSetSingle, close_client
from tests import unittest


class TestGridFSFile(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Start up a replica set and connect to it
        cls.repl_set = ReplicaSetSingle().start()
        cls.main_connection = cls.repl_set.client()

    @classmethod
    def tearDownClass(cls):
        close_client(cls.main_connection)
        cls.repl_set.stop()

    def setUp(self):
        # clear existing data
        self.main_connection.drop_database("test")
        self.collection = self.main_connection.test.fs
        self.fs = gridfs.GridFS(self.main_connection.test)

    def get_file(self, doc):
        return GridFSFile(self.collection, doc)

    def test_insert(self):
        def test_insert_file(data, filename, read_size):
            # insert file
            id = self.fs.put(data, filename=filename, encoding="utf8")
            doc = self.collection.files.find_one(id)
            f = self.get_file(doc)

            # test metadata
            self.assertEqual(id, f._id)
            self.assertEqual(filename, f.filename)

            # test data
            result = []
            while True:
                s = f.read(read_size)
                if len(s) > 0:
                    result.append(s.decode("utf8"))
                    if read_size >= 0:
                        self.assertLessEqual(len(s), read_size)
                else:
                    break
            result = "".join(result)
            self.assertEqual(f.length, len(result))
            self.assertEqual(data, result)

        # test with 1-chunk files
        test_insert_file("hello world", "hello.txt", -1)
        test_insert_file("hello world 2", "hello.txt", 10)
        test_insert_file("hello world 3", "hello.txt", 100)

        # test with multiple-chunk files
        size = 4 * 1024 * 1024
        bigger = "".join([chr(ord("a") + (n % 26)) for n in range(size)])
        test_insert_file(bigger, "bigger.txt", -1)
        test_insert_file(bigger, "bigger.txt", 1024)
        test_insert_file(bigger, "bigger.txt", 1024 * 1024)

    def test_missing_chunk(self):
        data = "test data"
        id = self.fs.put(data, encoding="utf8")
        doc = self.collection.files.find_one(id)
        f = self.get_file(doc)

        self.main_connection["test"]["fs.chunks"].delete_one({"files_id": id})

        self.assertRaises(errors.OperationFailed, f.read)


if __name__ == "__main__":
    unittest.main()



import gridfs

from mongo_connector import errors, util

wrap_exceptions = util.exception_wrapper(
    {gridfs.errors.CorruptGridFile: errors.OperationFailed}
)


class GridFSFile(object):
    @wrap_exceptions
    def __init__(self, collection, doc):
        self._id = doc["_id"]
        self.f = gridfs.GridOut(collection, file_document=doc)
        self.filename = self.f.filename
        self.length = self.f.length
        self.upload_date = self.f.upload_date
        self.md5 = self.f.md5

    def get_metadata(self):
        result = {"_id": self._id, "upload_date": self.upload_date, "md5": self.md5}
        if self.filename is not None:
            result["filename"] = self.filename
        return result

    def __len__(self):
        return self.length

    @wrap_exceptions
    def read(self, n=-1):
        return self.f.read(n)



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A class to serve as proxy for the target engine for testing.

Receives documents from the oplog worker threads and indexes them
into the backend.

Please look at the Solr and ElasticSearch doc manager classes for a sample
implementation with real systems.
"""

from threading import RLock

from mongo_connector import constants
from mongo_connector.errors import OperationFailed
from mongo_connector.doc_managers.doc_manager_base import DocManagerBase

__version__ = constants.__version__
"""DocManager Simulator version information

This is packaged with mongo-connector so it shares the same version.
Downstream DocManager implementations should add their package __version__
string here, for example:

__version__ = '0.1.0'
"""


class DocumentStore(dict):
    def __init__(self):
        self._lock = RLock()

    def __getitem__(self, key):
        with self._lock:
            return super(DocumentStore, self).__getitem__(key)

    def __setitem__(self, key, value):
        with self._lock:
            return super(DocumentStore, self).__setitem__(key, value)

    def __iter__(self):
        def __myiter__():
            with self._lock:
                for item in super(DocumentStore, self).__iter__():
                    yield item

        return __myiter__()


class Entry(object):
    def __init__(self, doc, ns, ts):
        self.doc, self.ns, self.ts = doc, ns, ts
        self._id = self.doc["_id"]

    @property
    def meta_dict(self):
        return {"_id": self._id, "ns": self.ns, "_ts": self.ts}

    @property
    def merged_dict(self):
        d = self.doc.copy()
        d.update(**self.meta_dict)
        return d

    def update(self, ns, ts):
        self.ns, self.ts = ns, ts


class DocManager(DocManagerBase):
    """BackendSimulator emulates both a target DocManager and a server.

    The DocManager class creates a connection to the backend engine and
    adds/removes documents, and in the case of rollback, searches for them.

    The reason for storing id/doc pairs as opposed to doc's is so that multiple
    updates to the same doc reflect the most up to date version as opposed to
    multiple, slightly different versions of a doc.
    """

    def __init__(
        self,
        url=None,
        unique_key="_id",
        auto_commit_interval=None,
        chunk_size=constants.DEFAULT_MAX_BULK,
        **kwargs
    ):
        """Creates a dictionary to hold document id keys mapped to the
        documents as values.
        """
        self.unique_key = unique_key
        self.auto_commit_interval = auto_commit_interval
        self.doc_dict = DocumentStore()
        self.url = url
        self.chunk_size = chunk_size
        self.kwargs = kwargs

    def stop(self):
        """Stops any running threads in the DocManager.
        """
        pass

    def update(self, document_id, update_spec, namespace, timestamp):
        """Apply updates given in update_spec to the document whose id
        matches that of doc.

        """
        document = self.doc_dict[document_id].doc
        updated = self.apply_update(document, update_spec)
        if "_id" in updated:
            updated.pop("_id")
        updated[self.unique_key] = document_id
        self.upsert(updated, namespace, timestamp)
        return updated

    def upsert(self, doc, namespace, timestamp):
        """Adds a document to the doc dict.
        """

        # Allow exceptions to be triggered (for testing purposes)
        if doc.get("_upsert_exception"):
            raise Exception("upsert exception")

        doc_id = doc["_id"]
        self.doc_dict[doc_id] = Entry(doc=doc, ns=namespace, ts=timestamp)

    def insert_file(self, f, namespace, timestamp):
        """Inserts a file to the doc dict.
        """
        doc = f.get_metadata()
        doc["content"] = f.read()
        self.doc_dict[f._id] = Entry(doc=doc, ns=namespace, ts=timestamp)

    def remove(self, document_id, namespace, timestamp):
        """Removes the document from the doc dict.
        """
        try:
            entry = self.doc_dict[document_id]
            entry.doc = None
            entry.update(namespace, timestamp)
        except KeyError:
            raise OperationFailed("Document does not exist: %s" % document_id)

    def search(self, start_ts, end_ts):
        """Searches through all documents and finds all documents that were
        modified or deleted within the range.

        Since we have very few documents in the doc dict when this is called,
        linear search is fine. This method is only used by rollbacks to query
        all the documents in the target engine within a certain timestamp
        window. The input will be two longs (converted from Bson timestamp)
        which specify the time range. The start_ts refers to the timestamp
        of the last oplog entry after a rollback. The end_ts is the timestamp
        of the last document committed to the backend.
        """
        for _id in self.doc_dict:
            entry = self.doc_dict[_id]
            if entry.ts <= end_ts or entry.ts >= start_ts:
                yield entry.meta_dict

    def commit(self):
        """Simply passes since we're not using an engine that needs commiting.
        """
        pass

    def get_last_doc(self):
        """Searches through the doc dict to find the document that was
        modified or deleted most recently."""
        return max(self.doc_dict.values(), key=lambda x: x.ts).meta_dict

    def handle_command(self, command_doc, namespace, timestamp):
        pass

    def _search(self):
        """Returns all documents in the doc dict.

        This function is not a part of the DocManager API, and is only used
        to simulate searching all documents from a backend.
        """
        results = []
        for _id in self.doc_dict:
            entry = self.doc_dict[_id]
            if entry.doc is not None:
                results.append(entry.merged_dict)
        return results

    def _delete(self):
        """Deletes all documents.

        This function is not a part of the DocManager API, and is only used
        to simulate deleting all documents from a backend.
        """
        self.doc_dict = {}



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import threading
import time

import bson
import pymongo
from pymongo.read_preferences import ReadPreference
from pymongo.write_concern import WriteConcern

sys.path[0:0] = [""]  # noqa

from mongo_connector.doc_managers.doc_manager_simulator import DocManager
from mongo_connector.locking_dict import LockingDict
from mongo_connector.namespace_config import NamespaceConfig
from mongo_connector.oplog_manager import OplogThread
from mongo_connector.test_utils import (
    assert_soon,
    close_client,
    ShardedCluster,
    ShardedClusterSingle,
)
from mongo_connector.util import retry_until_ok, bson_ts_to_long
from tests import unittest, SkipTest


class ShardedClusterTestCase(unittest.TestCase):
    def set_up_sharded_cluster(self, sharded_cluster_type):
        """ Initialize the cluster:

        Clean out the databases used by the tests
        Make connections to mongos, mongods
        Create and shard test collections
        Create OplogThreads
        """
        self.cluster = sharded_cluster_type().start()

        # Connection to mongos
        self.mongos_conn = self.cluster.client()

        # Connections to the shards
        self.shard1_conn = self.cluster.shards[0].client()
        self.shard2_conn = self.cluster.shards[1].client()

        # Wipe any test data
        self.mongos_conn["test"]["mcsharded"].drop()

        # Disable the balancer before creating the collection
        self.mongos_conn.config.settings.update_one(
            {"_id": "balancer"}, {"$set": {"stopped": True}}, upsert=True
        )

        # Create and shard the collection test.mcsharded on the "i" field
        self.mongos_conn["test"]["mcsharded"].create_index("i")
        self.mongos_conn.admin.command("enableSharding", "test")
        self.mongos_conn.admin.command(
            "shardCollection", "test.mcsharded", key={"i": 1}
        )

        # Pre-split the collection so that:
        # i < 1000            lives on shard1
        # i >= 1000           lives on shard2
        self.mongos_conn.admin.command(
            bson.SON([("split", "test.mcsharded"), ("middle", {"i": 1000})])
        )

        # Move chunks to their proper places
        try:
            self.mongos_conn["admin"].command(
                "moveChunk", "test.mcsharded", find={"i": 1}, to="demo-set-0"
            )
        except pymongo.errors.OperationFailure:
            pass
        try:
            self.mongos_conn["admin"].command(
                "moveChunk", "test.mcsharded", find={"i": 1000}, to="demo-set-1"
            )
        except pymongo.errors.OperationFailure:
            pass

        # Make sure chunks are distributed correctly
        self.mongos_conn["test"]["mcsharded"].insert_one({"i": 1})
        self.mongos_conn["test"]["mcsharded"].insert_one({"i": 1000})

        def chunks_moved():
            doc1 = self.shard1_conn.test.mcsharded.find_one()
            doc2 = self.shard2_conn.test.mcsharded.find_one()
            if None in (doc1, doc2):
                return False
            return doc1["i"] == 1 and doc2["i"] == 1000

        assert_soon(
            chunks_moved,
            max_tries=120,
            message="chunks not moved? doc1=%r, doc2=%r"
            % (
                self.shard1_conn.test.mcsharded.find_one(),
                self.shard2_conn.test.mcsharded.find_one(),
            ),
        )
        self.mongos_conn.test.mcsharded.delete_many({})

        # create a new oplog progress file
        try:
            os.unlink("oplog.timestamp")
        except OSError:
            pass
        open("oplog.timestamp", "w").close()

        # Oplog threads (oplog manager) for each shard
        doc_manager = DocManager()
        oplog_progress = LockingDict()
        namespace_config = NamespaceConfig(
            namespace_set=["test.mcsharded", "test.mcunsharded"]
        )
        self.opman1 = OplogThread(
            primary_client=self.shard1_conn,
            doc_managers=(doc_manager,),
            oplog_progress_dict=oplog_progress,
            namespace_config=namespace_config,
            mongos_client=self.mongos_conn,
        )
        self.opman2 = OplogThread(
            primary_client=self.shard2_conn,
            doc_managers=(doc_manager,),
            oplog_progress_dict=oplog_progress,
            namespace_config=namespace_config,
            mongos_client=self.mongos_conn,
        )

    def tearDown(self):
        try:
            self.opman1.join()
        except RuntimeError:
            pass  # thread may not have been started
        try:
            self.opman2.join()
        except RuntimeError:
            pass  # thread may not have been started
        close_client(self.mongos_conn)
        close_client(self.shard1_conn)
        close_client(self.shard2_conn)
        self.cluster.stop()


class TestOplogManagerShardedSingle(ShardedClusterTestCase):
    """Defines all test cases for OplogThreads running on a sharded
    cluster with single node replica sets.
    """

    def setUp(self):
        self.set_up_sharded_cluster(ShardedClusterSingle)

    def test_get_oplog_cursor(self):
        """Test the get_oplog_cursor method"""

        # timestamp = None
        cursor1 = self.opman1.get_oplog_cursor(None)
        oplog1 = self.shard1_conn["local"]["oplog.rs"].find({"op": {"$ne": "n"}})
        self.assertEqual(list(cursor1), list(oplog1))

        cursor2 = self.opman2.get_oplog_cursor(None)
        oplog2 = self.shard2_conn["local"]["oplog.rs"].find({"op": {"$ne": "n"}})
        self.assertEqual(list(cursor2), list(oplog2))

        # earliest entry is the only one at/after timestamp
        doc = {"ts": bson.Timestamp(1000, 0), "i": 1}
        self.mongos_conn["test"]["mcsharded"].insert_one(doc)
        latest_timestamp = self.opman1.get_last_oplog_timestamp()
        cursor = self.opman1.get_oplog_cursor(latest_timestamp)
        self.assertNotEqual(cursor, None)
        entries = list(cursor)
        self.assertEqual(len(entries), 1)
        next_entry_id = entries[0]["o"]["_id"]
        retrieved = self.mongos_conn.test.mcsharded.find_one(next_entry_id)
        self.assertEqual(retrieved, doc)

        # many entries before and after timestamp
        for i in range(2, 2002):
            self.mongos_conn["test"]["mcsharded"].insert_one({"i": i})
        oplog1 = self.shard1_conn["local"]["oplog.rs"].find(
            sort=[("ts", pymongo.ASCENDING)]
        )
        oplog2 = self.shard2_conn["local"]["oplog.rs"].find(
            sort=[("ts", pymongo.ASCENDING)]
        )

        # oplogs should have records for inserts performed, plus
        # various other messages
        oplog1_count = oplog1.count()
        oplog2_count = oplog2.count()
        self.assertGreaterEqual(oplog1_count, 998)
        self.assertGreaterEqual(oplog2_count, 1002)
        pivot1 = oplog1.skip(400).limit(-1)[0]
        pivot2 = oplog2.skip(400).limit(-1)[0]

        cursor1 = self.opman1.get_oplog_cursor(pivot1["ts"])
        cursor2 = self.opman2.get_oplog_cursor(pivot2["ts"])
        self.assertEqual(cursor1.count(), oplog1_count - 400)
        self.assertEqual(cursor2.count(), oplog2_count - 400)

    def test_get_last_oplog_timestamp(self):
        """Test the get_last_oplog_timestamp method"""

        # "empty" the oplog
        self.opman1.oplog = self.shard1_conn["test"]["emptycollection"]
        self.opman2.oplog = self.shard2_conn["test"]["emptycollection"]
        self.assertEqual(self.opman1.get_last_oplog_timestamp(), None)
        self.assertEqual(self.opman2.get_last_oplog_timestamp(), None)

        # Test non-empty oplog
        self.opman1.oplog = self.shard1_conn["local"]["oplog.rs"]
        self.opman2.oplog = self.shard2_conn["local"]["oplog.rs"]
        for i in range(1000):
            self.mongos_conn["test"]["mcsharded"].insert_one({"i": i + 500})
        oplog1 = self.shard1_conn["local"]["oplog.rs"]
        oplog1 = oplog1.find().sort("$natural", pymongo.DESCENDING).limit(-1)[0]
        oplog2 = self.shard2_conn["local"]["oplog.rs"]
        oplog2 = oplog2.find().sort("$natural", pymongo.DESCENDING).limit(-1)[0]
        self.assertEqual(self.opman1.get_last_oplog_timestamp(), oplog1["ts"])
        self.assertEqual(self.opman2.get_last_oplog_timestamp(), oplog2["ts"])

    def test_dump_collection(self):
        """Test the dump_collection method

        Cases:

        1. empty oplog
        2. non-empty oplog
        """

        # Test with empty oplog
        self.opman1.oplog = self.shard1_conn["test"]["emptycollection"]
        self.opman2.oplog = self.shard2_conn["test"]["emptycollection"]
        last_ts1 = self.opman1.dump_collection()
        last_ts2 = self.opman2.dump_collection()
        self.assertEqual(last_ts1, None)
        self.assertEqual(last_ts2, None)

        # Test with non-empty oplog
        self.opman1.oplog = self.shard1_conn["local"]["oplog.rs"]
        self.opman2.oplog = self.shard2_conn["local"]["oplog.rs"]
        for i in range(1000):
            self.mongos_conn["test"]["mcsharded"].insert_one({"i": i + 500})
        last_ts1 = self.opman1.get_last_oplog_timestamp()
        last_ts2 = self.opman2.get_last_oplog_timestamp()
        self.assertEqual(last_ts1, self.opman1.dump_collection())
        self.assertEqual(last_ts2, self.opman2.dump_collection())
        self.assertEqual(len(self.opman1.doc_managers[0]._search()), 1000)

    def test_init_cursor(self):
        """Test the init_cursor method

        Cases:

        1. no last checkpoint, no collection dump
        2. no last checkpoint, collection dump ok and stuff to dump
        3. no last checkpoint, nothing to dump, stuff in oplog
        4. no last checkpoint, nothing to dump, nothing in oplog
        5. no last checkpoint, no collection dump, stuff in oplog
        6. last checkpoint exists
        7. last checkpoint is behind
        """

        # N.B. these sub-cases build off of each other and cannot be re-ordered
        # without side-effects

        # No last checkpoint, no collection dump, nothing in oplog
        # "change oplog collection" to put nothing in oplog
        self.opman1.oplog = self.shard1_conn["test"]["emptycollection"]
        self.opman2.oplog = self.shard2_conn["test"]["emptycollection"]
        self.opman1.collection_dump = False
        self.opman2.collection_dump = False
        self.assertTrue(all(doc["op"] == "n" for doc in self.opman1.init_cursor()[0]))
        self.assertEqual(self.opman1.checkpoint, None)
        self.assertTrue(all(doc["op"] == "n" for doc in self.opman2.init_cursor()[0]))
        self.assertEqual(self.opman2.checkpoint, None)

        # No last checkpoint, empty collections, nothing in oplog
        self.opman1.collection_dump = self.opman2.collection_dump = True

        cursor, cursor_empty = self.opman1.init_cursor()
        self.assertEqual(cursor, None)
        self.assertTrue(cursor_empty)
        self.assertEqual(self.opman1.checkpoint, None)
        cursor, cursor_empty = self.opman2.init_cursor()
        self.assertEqual(cursor, None)
        self.assertTrue(cursor_empty)
        self.assertEqual(self.opman2.checkpoint, None)

        # No last checkpoint, empty collections, something in oplog
        self.opman1.oplog = self.shard1_conn["local"]["oplog.rs"]
        self.opman2.oplog = self.shard2_conn["local"]["oplog.rs"]
        oplog_startup_ts = self.opman2.get_last_oplog_timestamp()
        collection = self.mongos_conn["test"]["mcsharded"]
        collection.insert_one({"i": 1})
        collection.delete_one({"i": 1})
        time.sleep(3)
        last_ts1 = self.opman1.get_last_oplog_timestamp()
        cursor, cursor_empty = self.opman1.init_cursor()
        self.assertFalse(cursor_empty)
        self.assertEqual(self.opman1.checkpoint, last_ts1)
        self.assertEqual(self.opman1.read_last_checkpoint(), last_ts1)
        # init_cursor should point to startup message in shard2 oplog
        cursor, cursor_empty = self.opman2.init_cursor()
        self.assertFalse(cursor_empty)
        self.assertEqual(self.opman2.checkpoint, oplog_startup_ts)

        # No last checkpoint, no collection dump, stuff in oplog
        # If collection dump is false the checkpoint should not be set
        progress = LockingDict()
        self.opman1.oplog_progress = self.opman2.oplog_progress = progress
        self.opman1.collection_dump = self.opman2.collection_dump = False
        self.opman1.checkpoint = self.opman2.checkpoint = None
        collection.insert_one({"i": 1200})
        cursor, cursor_empty = self.opman1.init_cursor()
        self.assertFalse(cursor_empty)
        self.assertIsNone(self.opman1.checkpoint)
        self.assertEqual(next(cursor), next(self.opman1.get_oplog_cursor()))
        cursor, cursor_empty = self.opman2.init_cursor()
        self.assertFalse(cursor_empty)
        self.assertIsNone(self.opman2.checkpoint)
        for doc in cursor:
            last_doc = doc
        self.assertEqual(last_doc["o"]["i"], 1200)

        # Last checkpoint exists
        collection.insert_many([{"i": i + 500} for i in range(1000)])
        entry1 = list(self.shard1_conn["local"]["oplog.rs"].find(skip=200, limit=-2))
        entry2 = list(self.shard2_conn["local"]["oplog.rs"].find(skip=200, limit=-2))
        self.opman1.update_checkpoint(entry1[0]["ts"])
        self.opman2.update_checkpoint(entry2[0]["ts"])
        self.opman1.checkpoint = self.opman2.checkpoint = None
        cursor1, _ = self.opman1.init_cursor()
        cursor2, _ = self.opman2.init_cursor()
        self.assertEqual(entry1[1]["ts"], next(cursor1)["ts"])
        self.assertEqual(entry2[1]["ts"], next(cursor2)["ts"])
        self.assertEqual(self.opman1.checkpoint, entry1[0]["ts"])
        self.assertEqual(self.opman2.checkpoint, entry2[0]["ts"])
        self.assertEqual(self.opman1.read_last_checkpoint(), entry1[0]["ts"])
        self.assertEqual(self.opman2.read_last_checkpoint(), entry2[0]["ts"])

        # Last checkpoint is behind
        self.opman1.update_checkpoint(bson.Timestamp(1, 0))
        self.opman2.update_checkpoint(bson.Timestamp(1, 0))
        cursor, cursor_empty = self.opman1.init_cursor()
        self.assertTrue(cursor_empty)
        self.assertEqual(cursor, None)
        self.assertIsNotNone(self.opman1.checkpoint)
        cursor, cursor_empty = self.opman2.init_cursor()
        self.assertTrue(cursor_empty)
        self.assertEqual(cursor, None)
        self.assertIsNotNone(self.opman2.checkpoint)

    def test_with_chunk_migration(self):
        """Test that DocManagers have proper state after both a successful
        and an unsuccessful chunk migration
        """

        # Start replicating to dummy doc managers
        self.opman1.start()
        self.opman2.start()

        collection = self.mongos_conn["test"]["mcsharded"]
        for i in range(1000):
            collection.insert_one({"i": i + 500})
        # Assert current state of the mongoverse
        self.assertEqual(self.shard1_conn["test"]["mcsharded"].find().count(), 500)
        self.assertEqual(self.shard2_conn["test"]["mcsharded"].find().count(), 500)
        assert_soon(lambda: len(self.opman1.doc_managers[0]._search()) == 1000)

        # Test successful chunk move from shard 1 to shard 2
        self.mongos_conn["admin"].command(
            "moveChunk", "test.mcsharded", find={"i": 1}, to="demo-set-1"
        )

        # doc manager should still have all docs
        all_docs = self.opman1.doc_managers[0]._search()
        self.assertEqual(len(all_docs), 1000)
        for i, doc in enumerate(sorted(all_docs, key=lambda x: x["i"])):
            self.assertEqual(doc["i"], i + 500)

        # Mark the collection as "dropped". This will cause migration to fail.
        self.mongos_conn["config"]["collections"].update_one(
            {"_id": "test.mcsharded"}, {"$set": {"dropped": True}}
        )

        # Test unsuccessful chunk move from shard 2 to shard 1
        def fail_to_move_chunk():
            self.mongos_conn["admin"].command(
                "moveChunk", "test.mcsharded", find={"i": 1}, to="demo-set-0"
            )

        self.assertRaises(pymongo.errors.OperationFailure, fail_to_move_chunk)
        # doc manager should still have all docs
        all_docs = self.opman1.doc_managers[0]._search()
        self.assertEqual(len(all_docs), 1000)
        for i, doc in enumerate(sorted(all_docs, key=lambda x: x["i"])):
            self.assertEqual(doc["i"], i + 500)

    def test_upgrade_oplog_progress(self):
        first_oplog_ts1 = self.opman1.oplog.find_one()["ts"]
        first_oplog_ts2 = self.opman2.oplog.find_one()["ts"]
        # Old format oplog progress file:
        progress = {
            str(self.opman1.oplog): bson_ts_to_long(first_oplog_ts1),
            str(self.opman2.oplog): bson_ts_to_long(first_oplog_ts2),
        }
        # Set up oplog managers to use the old format.
        oplog_progress = LockingDict()
        oplog_progress.dict = progress
        self.opman1.oplog_progress = oplog_progress
        self.opman2.oplog_progress = oplog_progress
        # Cause the oplog managers to update their checkpoints.
        self.opman1.update_checkpoint(first_oplog_ts1)
        self.opman2.update_checkpoint(first_oplog_ts2)
        # New format should be in place now.
        new_format = {
            self.opman1.replset_name: first_oplog_ts1,
            self.opman2.replset_name: first_oplog_ts2,
        }
        self.assertEqual(new_format, self.opman1.oplog_progress.get_dict())
        self.assertEqual(new_format, self.opman2.oplog_progress.get_dict())


class TestOplogManagerSharded(ShardedClusterTestCase):
    """Defines all test cases for OplogThreads running on a sharded
    cluster with three node replica sets.
    """

    def setUp(self):
        self.set_up_sharded_cluster(ShardedCluster)
        self.shard1_secondary_conn = self.cluster.shards[0].secondary.client(
            read_preference=ReadPreference.SECONDARY_PREFERRED
        )
        self.shard2_secondary_conn = self.cluster.shards[1].secondary.client(
            read_preference=ReadPreference.SECONDARY_PREFERRED
        )

    def tearDown(self):
        super(TestOplogManagerSharded, self).tearDown()
        close_client(self.shard1_secondary_conn)
        close_client(self.shard2_secondary_conn)

    def test_with_orphan_documents(self):
        """Test that DocManagers have proper state after a chunk migration
        that resuts in orphaned documents.
        """
        # Start replicating to dummy doc managers
        self.opman1.start()
        self.opman2.start()

        collection = self.mongos_conn["test"]["mcsharded"]
        collection.insert_many([{"i": i + 500} for i in range(1000)])
        # Assert current state of the mongoverse
        self.assertEqual(self.shard1_conn["test"]["mcsharded"].find().count(), 500)
        self.assertEqual(self.shard2_conn["test"]["mcsharded"].find().count(), 500)
        assert_soon(lambda: len(self.opman1.doc_managers[0]._search()) == 1000)

        # Stop replication using the 'rsSyncApplyStop' failpoint.
        # Note: this requires secondaries to ensure the subsequent moveChunk
        # command does not complete.
        self.shard1_conn.admin.command(
            "configureFailPoint", "rsSyncApplyStop", mode="alwaysOn"
        )

        # Move a chunk from shard2 to shard1
        def move_chunk():
            try:
                self.mongos_conn["admin"].command(
                    "moveChunk", "test.mcsharded", find={"i": 1000}, to="demo-set-0"
                )
            except pymongo.errors.OperationFailure:
                pass

        # moveChunk will never complete, so use another thread to continue
        mover = threading.Thread(target=move_chunk)
        mover.start()

        # wait for documents to start moving to shard 1
        assert_soon(lambda: self.shard1_conn.test.mcsharded.count() > 500)

        # Get opid for moveChunk command
        operations = self.mongos_conn.test.current_op()
        opid = None
        for op in operations["inprog"]:
            if op.get("query", {}).get("moveChunk"):
                opid = op["opid"]

        if opid is None:
            raise SkipTest(
                "could not find moveChunk operation, cannot test " "failed moveChunk"
            )
        # Kill moveChunk with the opid

        if self.mongos_conn.server_info()["versionArray"][:3] >= [3, 1, 2]:
            self.mongos_conn.admin.command("killOp", op=opid)
        else:
            self.mongos_conn["test"]["$cmd.sys.killop"].find_one({"op": opid})

        # Mongo Connector should not become confused by unsuccessful chunk move
        docs = self.opman1.doc_managers[0]._search()
        self.assertEqual(len(docs), 1000)
        self.assertEqual(sorted(d["i"] for d in docs), list(range(500, 1500)))
        self.shard1_conn.admin.command(
            "configureFailPoint", "rsSyncApplyStop", mode="off"
        )
        # cleanup
        mover.join()

    def test_rollback(self):
        """Test the rollback method in a sharded environment

        Cases:
        1. Documents on both shards, rollback on one shard
        2. Documents on both shards, rollback on both shards

        """

        self.opman1.start()
        self.opman2.start()

        # Insert first documents while primaries are up
        db_main = self.mongos_conn["test"]["mcsharded"]
        db_main2 = db_main.with_options(write_concern=WriteConcern(w=2))
        db_main2.insert_one({"i": 0})
        db_main2.insert_one({"i": 1000})
        self.assertEqual(self.shard1_conn["test"]["mcsharded"].count(), 1)
        self.assertEqual(self.shard2_conn["test"]["mcsharded"].count(), 1)

        # Case 1: only one primary goes down, shard1 in this case
        self.cluster.shards[0].primary.stop(destroy=False)

        # Wait for the secondary to be promoted
        shard1_secondary_admin = self.shard1_secondary_conn["admin"]
        assert_soon(lambda: shard1_secondary_admin.command("isMaster")["ismaster"])

        # Insert another document. This will be rolled back later
        def insert_one(doc):
            if not db_main.find_one(doc):
                return db_main.insert_one(doc)
            return True

        assert_soon(
            lambda: retry_until_ok(insert_one, {"i": 1}),
            "could not insert into shard1 with one node down",
        )
        db_secondary1 = self.shard1_secondary_conn["test"]["mcsharded"]
        db_secondary2 = self.shard2_secondary_conn["test"]["mcsharded"]
        self.assertEqual(db_secondary1.count(), 2)

        # Wait for replication on the doc manager
        # Note that both OplogThreads share the same doc manager
        def c():
            return len(self.opman1.doc_managers[0]._search()) == 3

        assert_soon(c, "not all writes were replicated to doc manager", max_tries=120)

        # Kill the new primary
        self.cluster.shards[0].secondary.stop(destroy=False)

        # Start both servers back up
        self.cluster.shards[0].primary.start()
        primary_admin = self.shard1_conn["admin"]

        def c():
            return primary_admin.command("isMaster")["ismaster"]

        assert_soon(lambda: retry_until_ok(c))
        self.cluster.shards[0].secondary.start()
        secondary_admin = self.shard1_secondary_conn["admin"]

        def c():
            return secondary_admin.command("replSetGetStatus")["myState"] == 2

        assert_soon(c)
        query = {"i": {"$lt": 1000}}
        assert_soon(lambda: retry_until_ok(db_main.find(query).count) > 0)

        # Only first document should exist in MongoDB
        self.assertEqual(db_main.find(query).count(), 1)
        self.assertEqual(db_main.find_one(query)["i"], 0)

        def check_docman_rollback():
            docman_docs = [
                d for d in self.opman1.doc_managers[0]._search() if d["i"] < 1000
            ]
            return len(docman_docs) == 1 and docman_docs[0]["i"] == 0

        assert_soon(check_docman_rollback, "doc manager did not roll back")

        # Wait for previous rollback to complete.
        # Insert/delete one document to jump-start replication to secondaries
        # in MongoDB 3.x.
        db_main.insert_one({"i": -1})
        db_main.delete_one({"i": -1})

        def rollback_done():
            secondary1_count = retry_until_ok(db_secondary1.count)
            secondary2_count = retry_until_ok(db_secondary2.count)
            return (1, 1) == (secondary1_count, secondary2_count)

        assert_soon(
            rollback_done, "rollback never replicated to one or more secondaries"
        )

        ##############################

        # Case 2: Primaries on both shards go down
        self.cluster.shards[0].primary.stop(destroy=False)
        self.cluster.shards[1].primary.stop(destroy=False)

        # Wait for the secondaries to be promoted
        shard1_secondary_admin = self.shard1_secondary_conn["admin"]
        shard2_secondary_admin = self.shard2_secondary_conn["admin"]
        assert_soon(lambda: shard1_secondary_admin.command("isMaster")["ismaster"])
        assert_soon(lambda: shard2_secondary_admin.command("isMaster")["ismaster"])

        # Insert another document on each shard which will be rolled back later
        assert_soon(
            lambda: retry_until_ok(insert_one, {"i": 1}),
            "could not insert into shard1 with one node down",
        )
        self.assertEqual(db_secondary1.count(), 2)
        assert_soon(
            lambda: retry_until_ok(insert_one, {"i": 1001}),
            "could not insert into shard2 with one node down",
        )
        self.assertEqual(db_secondary2.count(), 2)

        # Wait for replication on the doc manager

        def c():
            return len(self.opman1.doc_managers[0]._search()) == 4

        assert_soon(c, "not all writes were replicated to doc manager")

        # Kill the new primaries
        self.cluster.shards[0].secondary.stop(destroy=False)
        self.cluster.shards[1].secondary.stop(destroy=False)

        # Start the servers back up...
        # Shard 1
        self.cluster.shards[0].primary.start()

        def c():
            return self.shard1_conn["admin"].command("isMaster")["ismaster"]

        assert_soon(lambda: retry_until_ok(c))
        self.cluster.shards[0].secondary.start()
        secondary_admin = self.shard1_secondary_conn["admin"]

        def c():
            return secondary_admin.command("replSetGetStatus")["myState"] == 2

        assert_soon(c)
        # Shard 2
        self.cluster.shards[1].primary.start()

        def c():
            return self.shard2_conn["admin"].command("isMaster")["ismaster"]

        assert_soon(lambda: retry_until_ok(c))
        self.cluster.shards[1].secondary.start()
        secondary_admin = self.shard2_secondary_conn["admin"]

        def c():
            return secondary_admin.command("replSetGetStatus")["myState"] == 2

        assert_soon(c)

        # Wait for the shards to come online
        assert_soon(lambda: retry_until_ok(db_main.find(query).count) > 0)
        query2 = {"i": {"$gte": 1000}}
        assert_soon(lambda: retry_until_ok(db_main.find(query2).count) > 0)

        # Only first documents should exist in MongoDB
        self.assertEqual(db_main.find(query).count(), 1)
        self.assertEqual(db_main.find_one(query)["i"], 0)
        self.assertEqual(db_main.find(query2).count(), 1)
        self.assertEqual(db_main.find_one(query2)["i"], 1000)

        # Same should hold for the doc manager
        assert_soon(lambda: len(self.opman1.doc_managers[0]._search()) == 2)
        i_values = [d["i"] for d in self.opman1.doc_managers[0]._search()]
        self.assertIn(0, i_values)
        self.assertIn(1000, i_values)


if __name__ == "__main__":
    unittest.main()



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests methods in util.py
"""
import sys

from bson import timestamp
from pymongo import errors

sys.path[0:0] = [""]  # noqa

from mongo_connector.util import bson_ts_to_long, long_to_bson_ts, retry_until_ok
from tests import unittest


def err_func(first_error):
    """Helper function for retry_until_ok test
    """
    err_func.counter += 1
    if err_func.counter == 3:
        return True
    elif err_func.counter == 2:
        raise TypeError
    else:
        raise first_error


err_func.counter = 0


class TestUtil(unittest.TestCase):
    """ Tests the utils
    """

    def setUp(self):
        err_func.counter = 0

    def test_bson_ts_to_long(self):
        """Test bson_ts_to_long and long_to_bson_ts
        """

        tstamp = timestamp.Timestamp(0x12345678, 0x90ABCDEF)

        self.assertEqual(0x1234567890ABCDEF, bson_ts_to_long(tstamp))
        self.assertEqual(long_to_bson_ts(0x1234567890ABCDEF), tstamp)

    def test_retry_until_ok(self):
        """Test retry_until_ok
        """
        self.assertTrue(retry_until_ok(err_func, errors.ConnectionFailure()))
        self.assertEqual(err_func.counter, 3)

    def test_retry_until_ok_operation_failure(self):
        """Test retry_until_ok retries on PyMongo OperationFailure.
        """
        self.assertTrue(retry_until_ok(err_func, errors.OperationFailure("")))
        self.assertEqual(err_func.counter, 3)

    def test_retry_until_ok_authorization(self):
        """Test retry_until_ok does not mask authorization failures.
        """
        with self.assertRaises(errors.OperationFailure):
            retry_until_ok(err_func, errors.OperationFailure("", 13, None))
        self.assertEqual(err_func.counter, 1)

    def test_retry_until_ok_authorization_mongodb_24(self):
        """Test retry_until_ok does not mask authorization failures in
        MongoDB 2.4.
        """
        with self.assertRaises(errors.OperationFailure):
            retry_until_ok(
                err_func,
                errors.OperationFailure("", details={"errmsg": "unauthorized"}),
            )
        self.assertEqual(err_func.counter, 1)

    def test_retry_until_ok_runtime_error(self):
        """Test retry_until_ok does not mask RuntimeErrors.
        """
        with self.assertRaises(RuntimeError):
            retry_until_ok(err_func, RuntimeError)
        self.assertEqual(err_func.counter, 1)


if __name__ == "__main__":

    unittest.main()



# Copyright 2016 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test include and exclude fields
"""
import sys

sys.path[0:0] = [""]  # noqa

from mongo_connector.doc_managers.doc_manager_simulator import DocManager
from mongo_connector.locking_dict import LockingDict
from mongo_connector.oplog_manager import OplogThread
from mongo_connector.namespace_config import NamespaceConfig
from mongo_connector.test_utils import assert_soon, close_client, ReplicaSetSingle
from tests import unittest


class TestFilterFields(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.repl_set = ReplicaSetSingle().start()
        cls.primary_conn = cls.repl_set.client()
        cls.oplog_coll = cls.primary_conn.local["oplog.rs"]

    @classmethod
    def tearDownClass(cls):
        cls.primary_conn.drop_database("test")
        close_client(cls.primary_conn)
        cls.repl_set.stop()

    def setUp(self):
        self.namespace_config = NamespaceConfig()
        self.opman = OplogThread(
            primary_client=self.primary_conn,
            doc_managers=(DocManager(),),
            oplog_progress_dict=LockingDict(),
            namespace_config=self.namespace_config,
        )

    def tearDown(self):
        try:
            self.opman.join()
        except RuntimeError:
            # OplogThread may not have been started
            pass

    def reset_include_fields(self, fields):
        self.opman.namespace_config = NamespaceConfig(include_fields=fields)

    def reset_exclude_fields(self, fields):
        self.opman.namespace_config = NamespaceConfig(exclude_fields=fields)

    def test_filter_fields(self):
        docman = self.opman.doc_managers[0]
        conn = self.opman.primary_client

        include_fields = ["a", "b", "c"]
        exclude_fields = ["d", "e", "f"]

        # Set fields to care about
        self.reset_include_fields(include_fields)
        # Documents have more than just these fields
        doc = {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6, "_id": 1}
        db = conn["test"]["test"]
        db.insert_one(doc)
        assert_soon(lambda: db.count() == 1)
        self.opman.dump_collection()

        result = docman._search()[0]
        keys = result.keys()
        for inc, exc in zip(include_fields, exclude_fields):
            self.assertIn(inc, keys)
            self.assertNotIn(exc, keys)

    def test_filter_exclude_oplog_entry(self):
        # Test oplog entries: these are callables, since
        # filter_oplog_entry modifies the oplog entry in-place
        def insert_op():
            return {"op": "i", "o": {"_id": 0, "a": 1, "b": 2, "c": 3}}

        def update_op():
            return {
                "op": "u",
                "o": {"$set": {"a": 4, "b": 5}, "$unset": {"c": True}},
                "o2": {"_id": 1},
            }

        def filter_doc(document, fields):
            if fields and "_id" in fields:
                fields.remove("_id")
            return self.opman.filter_oplog_entry(document, exclude_fields=fields)

        # Case 0: insert op, no fields provided
        filtered = filter_doc(insert_op(), None)
        self.assertEqual(filtered, insert_op())

        # Case 1: insert op, fields provided
        filtered = filter_doc(insert_op(), ["c"])
        self.assertEqual(filtered["o"], {"_id": 0, "a": 1, "b": 2})

        # Case 2: insert op, fields provided, doc becomes empty except for _id
        filtered = filter_doc(insert_op(), ["a", "b", "c"])
        self.assertEqual(filtered["o"], {"_id": 0})

        # Case 3: update op, no fields provided
        filtered = filter_doc(update_op(), None)
        self.assertEqual(filtered, update_op())

        # Case 4: update op, fields provided
        filtered = filter_doc(update_op(), ["b"])
        self.assertNotIn("b", filtered["o"]["$set"])
        self.assertIn("a", filtered["o"]["$set"])
        self.assertEqual(filtered["o"]["$unset"], update_op()["o"]["$unset"])

        # Case 5: update op, fields provided, empty $set
        filtered = filter_doc(update_op(), ["a", "b"])
        self.assertNotIn("$set", filtered["o"])
        self.assertEqual(filtered["o"]["$unset"], update_op()["o"]["$unset"])

        # Case 6: update op, fields provided, empty $unset
        filtered = filter_doc(update_op(), ["c"])
        self.assertNotIn("$unset", filtered["o"])
        self.assertEqual(filtered["o"]["$set"], update_op()["o"]["$set"])

        # Case 7: update op, fields provided, entry is nullified
        filtered = filter_doc(update_op(), ["a", "b", "c"])
        self.assertEqual(filtered, None)

        # Case 8: update op, fields provided, replacement
        filtered = filter_doc(
            {"op": "u", "o": {"a": 1, "b": 2, "c": 3, "d": 4}}, ["d", "e", "f"]
        )
        self.assertEqual(filtered, {"op": "u", "o": {"a": 1, "b": 2, "c": 3}})

    def test_filter_oplog_entry(self):
        # Test oplog entries: these are callables, since
        # filter_oplog_entry modifies the oplog entry in-place
        def insert_op():
            return {"op": "i", "o": {"_id": 0, "a": 1, "b": 2, "c": 3}}

        def update_op():
            return {
                "op": "u",
                "o": {"$set": {"a": 4, "b": 5}, "$unset": {"c": True}},
                "o2": {"_id": 1},
            }

        def filter_doc(document, fields):
            if fields and "_id" not in fields:
                fields.append("_id")
            return self.opman.filter_oplog_entry(document, include_fields=fields)

        # Case 0: insert op, no fields provided
        filtered = filter_doc(insert_op(), None)
        self.assertEqual(filtered, insert_op())

        # Case 1: insert op, fields provided
        filtered = filter_doc(insert_op(), ["a", "b"])
        self.assertEqual(filtered["o"], {"_id": 0, "a": 1, "b": 2})

        # Case 2: insert op, fields provided, doc becomes empty except for _id
        filtered = filter_doc(insert_op(), ["d", "e", "f"])
        self.assertEqual(filtered["o"], {"_id": 0})

        # Case 3: update op, no fields provided
        filtered = filter_doc(update_op(), None)
        self.assertEqual(filtered, update_op())

        # Case 4: update op, fields provided
        filtered = filter_doc(update_op(), ["a", "c"])
        self.assertNotIn("b", filtered["o"]["$set"])
        self.assertIn("a", filtered["o"]["$set"])
        self.assertEqual(filtered["o"]["$unset"], update_op()["o"]["$unset"])

        # Case 5: update op, fields provided, empty $set
        filtered = filter_doc(update_op(), ["c"])
        self.assertNotIn("$set", filtered["o"])
        self.assertEqual(filtered["o"]["$unset"], update_op()["o"]["$unset"])

        # Case 6: update op, fields provided, empty $unset
        filtered = filter_doc(update_op(), ["a", "b"])
        self.assertNotIn("$unset", filtered["o"])
        self.assertEqual(filtered["o"]["$set"], update_op()["o"]["$set"])

        # Case 7: update op, fields provided, entry is nullified
        filtered = filter_doc(update_op(), ["d", "e", "f"])
        self.assertEqual(filtered, None)

        # Case 8: update op, fields provided, replacement
        filtered = filter_doc(
            {"op": "u", "o": {"a": 1, "b": 2, "c": 3, "d": 4}}, ["a", "b", "c"]
        )
        self.assertEqual(filtered, {"op": "u", "o": {"a": 1, "b": 2, "c": 3}})

    def test_nested_fields(self):
        def check_nested(document, fields, filtered_document, op="i"):
            if "_id" not in fields:
                fields.append("_id")
            filtered_result = self.opman.filter_oplog_entry(
                {"op": op, "o": document}, include_fields=fields
            )
            if filtered_result is not None:
                filtered_result = filtered_result["o"]
            self.assertEqual(filtered_result, filtered_document)

        document = {"name": "Han Solo", "a": {"b": {}}}
        fields = ["name", "a.b.c"]
        filtered_document = {"name": "Han Solo"}
        check_nested(document, fields, filtered_document)

        document = {"a": {"b": {"c": 2, "e": 3}, "e": 5}, "b": 2, "c": {"g": 1}}
        fields = ["a.b.c", "a.e"]
        filtered_document = {"a": {"b": {"c": 2}, "e": 5}}
        check_nested(document, fields, filtered_document)

        document = {
            "a": {"b": {"c": 2, "e": 3}, "e": 5},
            "b": 2,
            "c": {"g": 1},
            "_id": 1,
        }
        fields = ["a.b.c", "a.e"]
        filtered_document = {"a": {"b": {"c": 2}, "e": 5}, "_id": 1}
        check_nested(document, fields, filtered_document)

        document = {"a": {"b": {"c": {"d": 1}}}, "-a": {"-b": {"-c": 2}}}
        fields = ["a.b", "-a"]
        filtered_document = document.copy()
        check_nested(document, fields, filtered_document)

        document = {"a": {"b": {"c": {"d": 1}}}, "-a": {"-b": {"-c": 2}}}
        fields = ["a", "-a.-b"]
        filtered_document = document.copy()
        check_nested(document, fields, filtered_document)
        document = {"a": {"b": {"c": {"d": 1}}}, "-a": {"-b": {"-c": 2}}, "_id": 1}

        fields = ["a.b", "-a"]
        filtered_document = document.copy()
        check_nested(document, fields, filtered_document)
        fields = ["a", "-a.-b"]
        check_nested(document, fields, filtered_document)

        document = {"test": 1}
        fields = ["doesnt_exist"]
        filtered_document = {}
        check_nested(document, fields, filtered_document)

        document = {"a": {"b": 1}, "b": {"a": 1}}
        fields = ["a.b", "b.a"]
        filtered_document = document.copy()
        check_nested(document, fields, filtered_document)

        document = {"a": {"b": {"a": {"b": 1}}}, "c": {"a": {"b": 1}}}
        fields = ["a.b"]
        filtered_document = {"a": {"b": {"a": {"b": 1}}}}
        check_nested(document, fields, filtered_document)

        document = {"name": "anna", "name_of_cat": "pushkin"}
        fields = ["name"]
        filtered_document = {"name": "anna"}
        check_nested(document, fields, filtered_document)

        update = {"$set": {"a.b": 1, "a.c": 3, "b": 2, "c": {"b": 3}}}
        fields = ["a", "c"]
        filtered_update = {"$set": {"a.b": 1, "a.c": 3, "c": {"b": 3}}}
        check_nested(update, fields, filtered_update, op="u")

        update = {"$set": {"a.b": {"c": 3, "d": 1}, "a.e": 1, "a.f": 2}}
        fields = ["a.b.c", "a.e"]
        filtered_update = {"$set": {"a.b": {"c": 3}, "a.e": 1}}
        check_nested(update, fields, filtered_update, op="u")

        update = {"$set": {"a.b.1": 1, "a.b.2": 2, "b": 3}}
        fields = ["a.b"]
        filtered_update = {"$set": {"a.b.1": 1, "a.b.2": 2}}
        check_nested(update, fields, filtered_update, op="u")

        update = {"$set": {"a.b": {"c": 3, "d": 1}, "a.e": 1}}
        fields = ["a.b.e"]
        filtered_update = None
        check_nested(update, fields, filtered_update, op="u")

    def test_nested_exclude_fields(self):
        def check_nested(document, exclude_fields, filtered_document, op="i"):
            if "_id" in exclude_fields:
                exclude_fields.remove("_id")
            filtered_result = self.opman.filter_oplog_entry(
                {"op": op, "o": document}, exclude_fields=exclude_fields
            )
            if filtered_result is not None:
                filtered_result = filtered_result["o"]
            self.assertEqual(filtered_result, filtered_document)

        document = {"a": {"b": {"c": {"d": 0, "e": 1}}}}
        exclude_fields = ["a.b.c.d"]
        filtered_document = {"a": {"b": {"c": {"e": 1}}}}
        check_nested(document, exclude_fields, filtered_document)

        document = {"a": {"b": {"c": {"-a": 0, "d": {"e": {"f": 1}}}}}}
        exclude_fields = ["a.b.c.d.e.f"]
        filtered_document = {"a": {"b": {"c": {"-a": 0, "d": {"e": {}}}}}}
        check_nested(document, exclude_fields, filtered_document)

        document = {"a": 1}
        exclude_fields = ["a"]
        filtered_document = {}
        check_nested(document, exclude_fields, filtered_document)

        document = {"a": {"b": {"c": 2, "e": 3}, "e": 5}, "b": 2, "c": {"g": 1}}
        exclude_fields = ["a.b.c", "a.e"]
        filtered_document = {"a": {"b": {"e": 3}}, "b": 2, "c": {"g": 1}}
        check_nested(document, exclude_fields, filtered_document)

        document = {
            "a": {"b": {"c": 2, "e": 3}, "e": 5},
            "b": 2,
            "c": {"g": 1},
            "_id": 1,
        }
        exclude_fields = ["a.b.c", "a.e", "_id"]
        filtered_document = {"a": {"b": {"e": 3}}, "b": 2, "c": {"g": 1}, "_id": 1}
        check_nested(document, exclude_fields, filtered_document)

        document = {"a": {"b": {"c": {"d": 1}}}, "-a": {"-b": {"-c": 2}}}
        exclude_fields = ["a.b", "-a"]
        filtered_document = {"a": {}}
        check_nested(document, exclude_fields, filtered_document)

        document = {"a": {"b": {"c": {"d": 1}}}, "-a": {"-b": {"-c": 2}}}
        exclude_fields = ["a", "-a.-b"]
        filtered_document = {"-a": {}}
        check_nested(document, exclude_fields, filtered_document)

        document = {"a": {"b": {"c": {"d": 1}}}, "-a": {"-b": {"-c": 2}}, "_id": 1}
        exclude_fields = ["a.b", "-a"]
        filtered_document = {"_id": 1, "a": {}}
        check_nested(document, exclude_fields, filtered_document)

        document = {"test": 1}
        exclude_fields = ["doesnt_exist"]
        filtered_document = document.copy()
        check_nested(document, exclude_fields, filtered_document)

        document = {"test": 1}
        exclude_fields = ["test.doesnt_exist"]
        filtered_document = document.copy()
        check_nested(document, exclude_fields, filtered_document)

        document = {"a": {"b": 1}, "b": {"a": 1}}
        exclude_fields = ["a.b", "b.a"]
        filtered_document = {"a": {}, "b": {}}
        check_nested(document, exclude_fields, filtered_document)

        document = {"a": {"b": {"a": {"b": 1}}}, "c": {"a": {"b": 1}}}
        exclude_fields = ["a.b"]
        filtered_document = {"a": {}, "c": {"a": {"b": 1}}}
        check_nested(document, exclude_fields, filtered_document)

        document = {"name": "anna", "name_of_cat": "pushkin"}
        exclude_fields = ["name"]
        filtered_document = {"name_of_cat": "pushkin"}
        check_nested(document, exclude_fields, filtered_document)

        update = {"$set": {"a.b": 1, "a.c": 3, "b": 2, "c": {"b": 3}}}
        exclude_fields = ["a", "c"]
        filtered_update = {"$set": {"b": 2}}
        check_nested(update, exclude_fields, filtered_update, op="u")

        update = {"$set": {"a.b": {"c": 3, "d": 1}, "a.e": 1, "a.f": 2}}
        exclude_fields = ["a.b.c", "a.e"]
        filtered_update = {"$set": {"a.b": {"d": 1}, "a.f": 2}}
        check_nested(update, exclude_fields, filtered_update, op="u")

        update = {"$set": {"a.b": {"c": 3, "d": 1}, "a.e": 1}}
        exclude_fields = ["a.b.c", "a.b.d", "a.e"]
        filtered_update = {"$set": {"a.b": {}}}
        check_nested(update, exclude_fields, filtered_update, op="u")

        update = {"$set": {"a.b.1": 1, "a.b.2": 2, "b": 3}}
        exclude_fields = ["a.b"]
        filtered_update = {"$set": {"b": 3}}
        check_nested(update, exclude_fields, filtered_update, op="u")

        update = {"$set": {"a.b.c": 42, "d.e.f": 123, "g": 456}}
        exclude_fields = ["a.b", "d"]
        filtered_update = {"$set": {"g": 456}}
        check_nested(update, exclude_fields, filtered_update, op="u")

        update = {"$set": {"a.b": {"c": 3, "d": 1}, "a.e": 1}}
        exclude_fields = ["a.b", "a.e"]
        filtered_update = None
        check_nested(update, exclude_fields, filtered_update, op="u")


class TestFindFields(unittest.TestCase):
    def test_find_field(self):
        doc = {"a": {"b": {"c": 1}}}
        self.assertEqual(OplogThread._find_field("a", doc), [(["a"], doc["a"])])
        self.assertEqual(
            OplogThread._find_field("a.b", doc), [(["a", "b"], doc["a"]["b"])]
        )
        self.assertEqual(
            OplogThread._find_field("a.b.c", doc),
            [(["a", "b", "c"], doc["a"]["b"]["c"])],
        )
        self.assertEqual(OplogThread._find_field("x", doc), [])
        self.assertEqual(OplogThread._find_field("a.b.x", doc), [])

    def test_find_update_fields(self):
        doc = {"a": {"b": {"c": 1}}, "e.f": 1, "g.h": {"i": {"j": 1}}}
        self.assertEqual(OplogThread._find_update_fields("a", doc), [(["a"], doc["a"])])
        self.assertEqual(
            OplogThread._find_update_fields("a.b", doc), [(["a", "b"], doc["a"]["b"])]
        )
        self.assertEqual(
            OplogThread._find_update_fields("a.b.c", doc),
            [(["a", "b", "c"], doc["a"]["b"]["c"])],
        )
        self.assertEqual(OplogThread._find_update_fields("x", doc), [])
        self.assertEqual(OplogThread._find_update_fields("a.b.x", doc), [])
        self.assertEqual(
            OplogThread._find_update_fields("e.f", doc), [(["e.f"], doc["e.f"])]
        )
        self.assertEqual(
            OplogThread._find_update_fields("e", doc), [(["e.f"], doc["e.f"])]
        )
        self.assertEqual(
            OplogThread._find_update_fields("g.h.i.j", doc),
            [(["g.h", "i", "j"], doc["g.h"]["i"]["j"])],
        )

        # Test multiple matches
        doc = {"a.b": 1, "a.c": 2, "e.f.h": 3, "e.f.i": 4}
        matches = OplogThread._find_update_fields("a", doc)
        self.assertEqual(len(matches), 2)
        self.assertIn((["a.b"], doc["a.b"]), matches)
        self.assertIn((["a.c"], doc["a.c"]), matches)
        matches = OplogThread._find_update_fields("e.f", doc)
        self.assertEqual(len(matches), 2)
        self.assertIn((["e.f.h"], doc["e.f.h"]), matches)
        self.assertIn((["e.f.i"], doc["e.f.i"]), matches)

        # Test updates to array fields
        doc = {"a.b.1": 9, "a.b.3": 10, "a.b.4.c": 11}
        matches = OplogThread._find_update_fields("a.b", doc)
        self.assertEqual(len(matches), 3)
        self.assertIn((["a.b.1"], doc["a.b.1"]), matches)
        self.assertIn((["a.b.3"], doc["a.b.3"]), matches)
        self.assertIn((["a.b.4.c"], doc["a.b.4.c"]), matches)


if __name__ == "__main__":
    unittest.main()



# Copyright 2013-2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tails the oplog of a shard and returns entries
"""

import bson
import logging

try:
    import Queue as queue
except ImportError:
    import queue
import sys
import time
import threading

import pymongo

from pymongo import CursorType

from mongo_connector import errors, util
from mongo_connector.constants import DEFAULT_BATCH_SIZE
from mongo_connector.gridfs_file import GridFSFile
from mongo_connector.util import log_fatal_exceptions, retry_until_ok

LOG = logging.getLogger(__name__)


class ReplicationLagLogger(threading.Thread):
    """Thread that periodically logs the current replication lag.
    """

    def __init__(self, opman, interval):
        super(ReplicationLagLogger, self).__init__()
        self.opman = opman
        self.interval = interval
        self.daemon = True

    def log_replication_lag(self):
        checkpoint = self.opman.checkpoint
        if checkpoint is None:
            return
        newest_write = retry_until_ok(self.opman.get_last_oplog_timestamp)
        if newest_write < checkpoint:
            # OplogThread will perform a rollback, don't log anything
            return
        lag_secs = newest_write.time - checkpoint.time
        if lag_secs > 0:
            LOG.info(
                "OplogThread for replica set '%s' is %s seconds behind " "the oplog.",
                self.opman.replset_name,
                lag_secs,
            )
        else:
            lag_inc = newest_write.inc - checkpoint.inc
            if lag_inc > 0:
                LOG.info(
                    "OplogThread for replica set '%s' is %s entries "
                    "behind the oplog.",
                    self.opman.replset_name,
                    lag_inc,
                )
            else:
                LOG.info(
                    "OplogThread for replica set '%s' is up to date " "with the oplog.",
                    self.opman.replset_name,
                )

    def run(self):
        while self.opman.is_alive():
            self.log_replication_lag()
            time.sleep(self.interval)


class OplogThread(threading.Thread):
    """Thread that tails an oplog.

    Calls the appropriate method on DocManagers for each relevant oplog entry.
    """

    def __init__(
        self,
        primary_client,
        doc_managers,
        oplog_progress_dict,
        namespace_config,
        mongos_client=None,
        **kwargs
    ):
        super(OplogThread, self).__init__()

        self.batch_size = kwargs.get("batch_size", DEFAULT_BATCH_SIZE)

        # The connection to the primary for this replicaSet.
        self.primary_client = primary_client

        # The connection to the mongos, if there is one.
        self.mongos_client = mongos_client

        # Are we allowed to perform a collection dump?
        self.collection_dump = kwargs.get("collection_dump", True)

        self.only_dump = kwargs.get('only_dump', False)

        # The document manager for each target system.
        # These are the same for all threads.
        self.doc_managers = doc_managers

        # Boolean describing whether or not the thread is running.
        self.running = True

        # Stores the timestamp of the last oplog entry read.
        self.checkpoint = None

        # A dictionary that stores OplogThread/timestamp pairs.
        # Represents the last checkpoint for a OplogThread.
        self.oplog_progress = oplog_progress_dict

        # The namespace configuration
        self.namespace_config = namespace_config

        # Whether the collection dump gracefully handles exceptions
        self.continue_on_error = kwargs.get("continue_on_error", False)

        LOG.info("OplogThread: Initializing oplog thread")

        self.oplog = self.primary_client.local.oplog.rs
        self.replset_name = self.primary_client.admin.command("ismaster")["setName"]

        if not self.oplog.find_one():
            err_msg = "OplogThread: No oplog for thread:"
            LOG.warning("%s %s" % (err_msg, self.primary_client))

    def _should_skip_entry(self, entry):
        """Determine if this oplog entry should be skipped.

        This has the possible side effect of modifying the entry's namespace
        and filtering fields from updates and inserts.
        """
        # Don't replicate entries resulting from chunk moves
        if entry.get("fromMigrate"):
            return True, False

        # Ignore no-ops
        if entry["op"] == "n":
            return True, False
        ns = entry["ns"]

        if "." not in ns:
            return True, False
        coll = ns.split(".", 1)[1]

        # Ignore system collections
        if coll.startswith("system."):
            return True, False

        # Ignore GridFS chunks
        if coll.endswith(".chunks"):
            return True, False

        is_gridfs_file = False
        if coll.endswith(".files"):
            ns = ns[: -len(".files")]
            if self.namespace_config.gridfs_namespace(ns):
                is_gridfs_file = True
            else:
                return True, False

        # Commands should not be ignored, filtered, or renamed. Renaming is
        # handled by the DocManagers via the CommandHelper class.
        if coll == "$cmd":
            return False, False

        # Rename or filter out namespaces that are ignored keeping
        # included gridfs namespaces.
        namespace = self.namespace_config.lookup(ns)
        if namespace is None:
            LOG.debug(
                "OplogThread: Skipping oplog entry: "
                "'%s' is not in the namespace configuration." % (ns,)
            )
            return True, False

        # Update the namespace.
        entry["ns"] = namespace.dest_name

        # Take fields out of the oplog entry that shouldn't be replicated.
        # This may nullify the document if there's nothing to do.
        if not self.filter_oplog_entry(
            entry,
            include_fields=namespace.include_fields,
            exclude_fields=namespace.exclude_fields,
        ):
            return True, False
        return False, is_gridfs_file

    @log_fatal_exceptions
    def run(self):
        """Start the oplog worker.
        """
        ReplicationLagLogger(self, 30).start()
        LOG.debug("OplogThread: Run thread started")
        while self.running is True:
            LOG.debug("OplogThread: Getting cursor")
            cursor, cursor_empty = retry_until_ok(self.init_cursor)
            # we've fallen too far behind
            if cursor is None and self.checkpoint is not None:
                err_msg = "OplogThread: Last entry no longer in oplog"
                effect = "cannot recover!"
                LOG.error("%s %s %s" % (err_msg, effect, self.oplog))
                self.running = False
                continue

            if cursor_empty:
                LOG.debug(
                    "OplogThread: Last entry is the one we "
                    "already processed.  Up to date.  Sleeping."
                )
                time.sleep(1)
                continue

            last_ts = None
            remove_inc = 0
            upsert_inc = 0
            update_inc = 0
            try:
                LOG.debug("OplogThread: about to process new oplog entries")
                while cursor.alive and self.running:
                    LOG.debug(
                        "OplogThread: Cursor is still"
                        " alive and thread is still running."
                    )
                    for n, entry in enumerate(cursor):
                        # Break out if this thread should stop
                        if not self.running:
                            break

                        LOG.debug(
                            "OplogThread: Iterating through cursor,"
                            " document number in this cursor is %d" % n
                        )

                        skip, is_gridfs_file = self._should_skip_entry(entry)
                        if skip:
                            # update the last_ts on skipped entries to ensure
                            # our checkpoint does not fall off the oplog. This
                            # also prevents reprocessing skipped entries.
                            last_ts = entry["ts"]
                            continue

                        # Sync the current oplog operation
                        operation = entry["op"]
                        ns = entry["ns"]
                        timestamp = util.bson_ts_to_long(entry["ts"])
                        for docman in self.doc_managers:
                            try:
                                LOG.debug(
                                    "OplogThread: Operation for this "
                                    "entry is %s" % str(operation)
                                )

                                # Remove
                                if operation == "d":
                                    docman.remove(entry["o"]["_id"], ns, timestamp)
                                    remove_inc += 1

                                # Insert
                                elif operation == "i":  # Insert
                                    # Retrieve inserted document from
                                    # 'o' field in oplog record
                                    doc = entry.get("o")
                                    # Extract timestamp and namespace
                                    if is_gridfs_file:
                                        db, coll = ns.split(".", 1)
                                        gridfile = GridFSFile(
                                            self.primary_client[db][coll], doc
                                        )
                                        docman.insert_file(gridfile, ns, timestamp)
                                    else:
                                        docman.upsert(doc, ns, timestamp)
                                    upsert_inc += 1

                                # Update
                                elif operation == "u":
                                    docman.update(
                                        entry["o2"]["_id"], entry["o"], ns, timestamp
                                    )
                                    update_inc += 1

                                # Command
                                elif operation == "c":
                                    # use unmapped namespace
                                    doc = entry.get("o")
                                    docman.handle_command(doc, entry["ns"], timestamp)

                            except errors.OperationFailed:
                                LOG.exception(
                                    "Unable to process oplog document %r" % entry
                                )
                            except errors.ConnectionFailed:
                                LOG.exception(
                                    "Connection failed while processing oplog "
                                    "document %r" % entry
                                )

                        if (remove_inc + upsert_inc + update_inc) % 1000 == 0:
                            LOG.debug(
                                "OplogThread: Documents removed: %d, "
                                "inserted: %d, updated: %d so far"
                                % (remove_inc, upsert_inc, update_inc)
                            )

                        LOG.debug("OplogThread: Doc is processed.")

                        last_ts = entry["ts"]

                        # update timestamp per batch size
                        # n % -1 (default for self.batch_size) == 0 for all n
                        if n % self.batch_size == 1:
                            self.update_checkpoint(last_ts)
                            last_ts = None

                    # update timestamp after running through oplog
                    if last_ts is not None:
                        LOG.debug(
                            "OplogThread: updating checkpoint after "
                            "processing new oplog entries"
                        )
                        self.update_checkpoint(last_ts)

            except (
                pymongo.errors.AutoReconnect,
                pymongo.errors.OperationFailure,
                pymongo.errors.ConfigurationError,
            ):
                LOG.exception(
                    "Cursor closed due to an exception. " "Will attempt to reconnect."
                )

            # update timestamp before attempting to reconnect to MongoDB,
            # after being join()'ed, or if the cursor closes
            if last_ts is not None:
                LOG.debug(
                    "OplogThread: updating checkpoint after an "
                    "Exception, cursor closing, or join() on this"
                    "thread."
                )
                self.update_checkpoint(last_ts)

            LOG.debug(
                "OplogThread: Sleeping. Documents removed: %d, "
                "upserted: %d, updated: %d" % (remove_inc, upsert_inc, update_inc)
            )
            time.sleep(2)

    def join(self):
        """Stop this thread from managing the oplog.
        """
        LOG.debug("OplogThread: exiting due to join call.")
        self.running = False
        threading.Thread.join(self)

    @classmethod
    def _find_field(cls, field, doc):
        """Find the field in the document which matches the given field.

        The field may be in dot notation, eg "a.b.c". Returns a list with
        a single tuple (path, field_value) or the empty list if the field
        is not present.
        """
        path = field.split(".")
        try:
            for key in path:
                doc = doc[key]
            return [(path, doc)]
        except (KeyError, TypeError):
            return []

    @classmethod
    def _find_update_fields(cls, field, doc):
        """Find the fields in the update document which match the given field.

        Both the field and the top level keys in the doc may be in dot
        notation, eg "a.b.c". Returns a list of tuples (path, field_value) or
        the empty list if the field is not present.
        """

        def find_partial_matches():
            for key in doc:
                if len(key) > len(field):
                    # Handle case where field is a prefix of key, eg field is
                    # 'a' and key is 'a.b'.
                    if key.startswith(field) and key[len(field)] == ".":
                        yield [key], doc[key]
                        # Continue searching, there may be multiple matches.
                        # For example, field 'a' should match 'a.b' and 'a.c'.
                elif len(key) < len(field):
                    # Handle case where key is a prefix of field, eg field is
                    # 'a.b' and key is 'a'.
                    if field.startswith(key) and field[len(key)] == ".":
                        # Search for the remaining part of the field
                        matched = cls._find_field(field[len(key) + 1 :], doc[key])
                        if matched:
                            # Add the top level key to the path.
                            match = matched[0]
                            match[0].insert(0, key)
                            yield match
                        # Stop searching, it's not possible for any other
                        # keys in the update doc to match this field.
                        return

        try:
            return [([field], doc[field])]
        except KeyError:
            # Field does not exactly match any key in the update doc.
            return list(find_partial_matches())

    def _pop_excluded_fields(self, doc, exclude_fields, update=False):
        # Remove all the fields that were passed in exclude_fields.
        find_fields = self._find_update_fields if update else self._find_field
        for field in exclude_fields:
            for path, _ in find_fields(field, doc):
                # Delete each matching field in the original document.
                temp_doc = doc
                for p in path[:-1]:
                    temp_doc = temp_doc[p]
                temp_doc.pop(path[-1])

        return doc  # Need this to be similar to copy_included_fields.

    def _copy_included_fields(self, doc, include_fields, update=False):
        new_doc = {}
        find_fields = self._find_update_fields if update else self._find_field
        for field in include_fields:
            for path, value in find_fields(field, doc):
                # Copy each matching field in the original document.
                temp_doc = new_doc
                for p in path[:-1]:
                    temp_doc = temp_doc.setdefault(p, {})
                temp_doc[path[-1]] = value

        return new_doc

    def filter_oplog_entry(self, entry, include_fields=None, exclude_fields=None):
        """Remove fields from an oplog entry that should not be replicated.

        NOTE: this does not support array indexing, for example 'a.b.2'"""
        if not include_fields and not exclude_fields:
            return entry
        elif include_fields:
            filter_fields = self._copy_included_fields
        else:
            filter_fields = self._pop_excluded_fields

        fields = include_fields or exclude_fields
        entry_o = entry["o"]
        # Version 3.6 of mongodb includes a $v,
        # see https://jira.mongodb.org/browse/SERVER-32240
        if "$v" in entry_o:
            entry_o.pop("$v")

        # 'i' indicates an insert. 'o' field is the doc to be inserted.
        if entry["op"] == "i":
            entry["o"] = filter_fields(entry_o, fields)
        # 'u' indicates an update. The 'o' field describes an update spec
        # if '$set' or '$unset' are present.
        elif entry["op"] == "u" and ("$set" in entry_o or "$unset" in entry_o):
            if "$set" in entry_o:
                entry["o"]["$set"] = filter_fields(entry_o["$set"], fields, update=True)
            if "$unset" in entry_o:
                entry["o"]["$unset"] = filter_fields(
                    entry_o["$unset"], fields, update=True
                )
            # not allowed to have empty $set/$unset, so remove if empty
            if "$set" in entry_o and not entry_o["$set"]:
                entry_o.pop("$set")
            if "$unset" in entry_o and not entry_o["$unset"]:
                entry_o.pop("$unset")
            if not entry_o:
                return None
        # 'u' indicates an update. The 'o' field is the replacement document
        # if no '$set' or '$unset' are present.
        elif entry["op"] == "u":
            entry["o"] = filter_fields(entry_o, fields)

        return entry

    def get_oplog_cursor(self, timestamp=None):
        """Get a cursor to the oplog after the given timestamp, excluding
        no-op entries.

        If no timestamp is specified, returns a cursor to the entire oplog.
        """
        query = {"op": {"$ne": "n"}}
        if timestamp is None:
            cursor = self.oplog.find(query, cursor_type=CursorType.TAILABLE_AWAIT)
        else:
            query["ts"] = {"$gte": timestamp}
            cursor = self.oplog.find(
                query, cursor_type=CursorType.TAILABLE_AWAIT, oplog_replay=True
            )
        return cursor

    def get_collection(self, namespace):
        """Get a pymongo collection from a namespace."""
        database, coll = namespace.split(".", 1)
        return self.primary_client[database][coll]

    def dump_collection(self):
        """Dumps collection into the target system.

        This method is called when we're initializing the cursor and have no
        configs i.e. when we're starting for the first time.
        """

        timestamp = retry_until_ok(self.get_last_oplog_timestamp)
        if timestamp is None:
            return None
        long_ts = util.bson_ts_to_long(timestamp)
        # Flag if this oplog thread was cancelled during the collection dump.
        # Use a list to workaround python scoping.
        dump_cancelled = [False]

        def get_all_ns():
            ns_set = []
            gridfs_ns_set = []
            db_list = self.namespace_config.get_included_databases()
            if not db_list:
                # Only use listDatabases when the configured databases are not
                # explicit.
                db_list = retry_until_ok(self.primary_client.database_names)
            for database in db_list:
                if database == "config" or database == "local":
                    continue
                coll_list = retry_until_ok(
                    self.primary_client[database].collection_names
                )
                for coll in coll_list:
                    # ignore system collections
                    if coll.startswith("system."):
                        continue
                    # ignore gridfs chunks collections
                    if coll.endswith(".chunks"):
                        continue
                    if coll.endswith(".files"):
                        namespace = "%s.%s" % (database, coll)
                        namespace = namespace[: -len(".files")]
                        if self.namespace_config.gridfs_namespace(namespace):
                            gridfs_ns_set.append(namespace)
                    else:
                        namespace = "%s.%s" % (database, coll)
                        if self.namespace_config.map_namespace(namespace):
                            ns_set.append(namespace)
            return ns_set, gridfs_ns_set

        dump_set, gridfs_dump_set = get_all_ns()

        LOG.debug("OplogThread: Dumping set of collections %s " % dump_set)

        def docs_to_dump(from_coll):
            last_id = None
            attempts = 0
            projection = self.namespace_config.projection(from_coll.full_name)
            # Loop to handle possible AutoReconnect
            while attempts < 60:
                if last_id is None:
                    cursor = retry_until_ok(
                        from_coll.find,
                        projection=projection,
                        sort=[("_id", pymongo.ASCENDING)],
                    )
                else:
                    cursor = retry_until_ok(
                        from_coll.find,
                        {"_id": {"$gt": last_id}},
                        projection=projection,
                        sort=[("_id", pymongo.ASCENDING)],
                    )
                try:
                    for doc in cursor:
                        if not self.running:
                            # Thread was joined while performing the
                            # collection dump.
                            dump_cancelled[0] = True
                            raise StopIteration
                        last_id = doc["_id"]
                        yield doc
                    break
                except (pymongo.errors.AutoReconnect, pymongo.errors.OperationFailure):
                    attempts += 1
                    time.sleep(1)

        def upsert_each(dm):
            num_failed = 0
            for namespace in dump_set:
                from_coll = self.get_collection(namespace)
                mapped_ns = self.namespace_config.map_namespace(namespace)
                total_docs = retry_until_ok(from_coll.count)
                num = None
                for num, doc in enumerate(docs_to_dump(from_coll)):
                    try:
                        dm.upsert(doc, mapped_ns, long_ts)
                    except Exception:
                        if self.continue_on_error:
                            LOG.exception("Could not upsert document: %r" % doc)
                            num_failed += 1
                        else:
                            raise
                    if num % 10000 == 0:
                        LOG.info(
                            "Upserted %d out of approximately %d docs "
                            "from collection '%s'",
                            num + 1,
                            total_docs,
                            namespace,
                        )
                if num is not None:
                    LOG.info(
                        "Upserted %d out of approximately %d docs from "
                        "collection '%s'",
                        num + 1,
                        total_docs,
                        namespace,
                    )
            if num_failed > 0:
                LOG.error("Failed to upsert %d docs" % num_failed)

        def upsert_all(dm):
            try:
                for namespace in dump_set:
                    from_coll = self.get_collection(namespace)
                    total_docs = retry_until_ok(from_coll.count)
                    mapped_ns = self.namespace_config.map_namespace(namespace)
                    LOG.info(
                        "Bulk upserting approximately %d docs from " "collection '%s'",
                        total_docs,
                        namespace,
                    )
                    dm.bulk_upsert(docs_to_dump(from_coll), mapped_ns, long_ts)
            except Exception:
                if self.continue_on_error:
                    LOG.exception(
                        "OplogThread: caught exception"
                        " during bulk upsert, re-upserting"
                        " documents serially"
                    )
                    upsert_each(dm)
                else:
                    raise

        def do_dump(dm, error_queue):
            try:
                LOG.debug(
                    "OplogThread: Using bulk upsert function for " "collection dump"
                )
                upsert_all(dm)

                if gridfs_dump_set:
                    LOG.info(
                        "OplogThread: dumping GridFS collections: %s", gridfs_dump_set
                    )

                # Dump GridFS files
                for gridfs_ns in gridfs_dump_set:
                    mongo_coll = self.get_collection(gridfs_ns)
                    from_coll = self.get_collection(gridfs_ns + ".files")
                    dest_ns = self.namespace_config.map_namespace(gridfs_ns)
                    for doc in docs_to_dump(from_coll):
                        gridfile = GridFSFile(mongo_coll, doc)
                        dm.insert_file(gridfile, dest_ns, long_ts)
            except Exception:
                # Likely exceptions:
                # pymongo.errors.OperationFailure,
                # mongo_connector.errors.ConnectionFailed
                # mongo_connector.errors.OperationFailed
                error_queue.put(sys.exc_info())

        # Extra threads (if any) that assist with collection dumps
        dumping_threads = []
        # Did the dump succeed for all target systems?
        dump_success = True
        # Holds any exceptions we can't recover from
        errors = queue.Queue()

        if len(self.doc_managers) == 1:
            do_dump(self.doc_managers[0], errors)
        else:
            # Slight performance gain breaking dump into separate
            # threads if > 1 replication target
            for dm in self.doc_managers:
                t = threading.Thread(target=do_dump, args=(dm, errors))
                dumping_threads.append(t)
                t.start()
            # cleanup
            for t in dumping_threads:
                t.join()

        # Print caught exceptions
        try:
            while True:
                LOG.critical(
                    "Exception during collection dump", exc_info=errors.get_nowait()
                )
                dump_success = False
        except queue.Empty:
            pass

        if not dump_success:
            err_msg = "OplogThread: Failed during dump collection"
            effect = "cannot recover!"
            LOG.error("%s %s %s" % (err_msg, effect, self.oplog))
            self.running = False
            return None

        if dump_cancelled[0]:
            LOG.warning(
                "Initial collection dump was interrupted. "
                "Will re-run the collection dump on next startup."
            )
            return None

        return timestamp

    def _get_oplog_timestamp(self, newest_entry):
        """Return the timestamp of the latest or earliest entry in the oplog.
        """
        sort_order = pymongo.DESCENDING if newest_entry else pymongo.ASCENDING
        curr = (
            self.oplog.find({"op": {"$ne": "n"}}).sort("$natural", sort_order).limit(-1)
        )

        try:
            ts = next(curr)["ts"]
        except StopIteration:
            LOG.debug("OplogThread: oplog is empty.")
            return None

        LOG.debug(
            "OplogThread: %s oplog entry has timestamp %s."
            % ("Newest" if newest_entry else "Oldest", ts)
        )
        return ts

    def get_oldest_oplog_timestamp(self):
        """Return the timestamp of the oldest entry in the oplog.
        """
        return self._get_oplog_timestamp(False)

    def get_last_oplog_timestamp(self):
        """Return the timestamp of the newest entry in the oplog.
        """
        return self._get_oplog_timestamp(True)

    def _cursor_empty(self, cursor):
        try:
            # Tailable cursors can not have singleBatch=True in MongoDB > 3.3
            next(cursor.clone().remove_option(CursorType.TAILABLE_AWAIT).limit(-1))
            return False
        except StopIteration:
            return True

    def init_cursor(self):
        """Position the cursor appropriately.

        The cursor is set to either the beginning of the oplog, or
        wherever it was last left off.

        Returns the cursor and True if the cursor is empty.
        """
        timestamp = self.read_last_checkpoint()

        if timestamp is None or self.only_dump:
            if self.collection_dump:
                # dump collection and update checkpoint
                timestamp = self.dump_collection()
                if self.only_dump:
                    LOG.info("Finished dump. Exiting.")
                    timestamp = None
                    self.running = False

                self.update_checkpoint(timestamp)
                if timestamp is None:
                    return None, True
            else:
                # Collection dump disabled:
                # Return cursor to beginning of oplog but do not set the
                # checkpoint. The checkpoint will be set after an operation
                # has been applied.
                cursor = self.get_oplog_cursor()
                return cursor, self._cursor_empty(cursor)

        cursor = self.get_oplog_cursor(timestamp)
        cursor_empty = self._cursor_empty(cursor)

        if cursor_empty:
            # rollback, update checkpoint, and retry
            LOG.debug("OplogThread: Initiating rollback from " "get_oplog_cursor")
            self.update_checkpoint(self.rollback())
            return self.init_cursor()

        first_oplog_entry = next(cursor)

        oldest_ts_long = util.bson_ts_to_long(self.get_oldest_oplog_timestamp())
        checkpoint_ts_long = util.bson_ts_to_long(timestamp)
        if checkpoint_ts_long < oldest_ts_long:
            # We've fallen behind, the checkpoint has fallen off the oplog
            return None, True

        cursor_ts_long = util.bson_ts_to_long(first_oplog_entry["ts"])
        if cursor_ts_long > checkpoint_ts_long:
            # The checkpoint is not present in this oplog and the oplog
            # did not rollover. This means that we connected to a new
            # primary which did not replicate the checkpoint and which has
            # new changes in its oplog for us to process.
            # rollback, update checkpoint, and retry
            LOG.debug(
                "OplogThread: Initiating rollback from "
                "get_oplog_cursor: new oplog entries found but "
                "checkpoint is not present"
            )
            self.update_checkpoint(self.rollback())
            return self.init_cursor()

        # first entry has been consumed
        return cursor, cursor_empty

    def update_checkpoint(self, checkpoint):
        """Store the current checkpoint in the oplog progress dictionary.
        """
        if checkpoint is not None and checkpoint != self.checkpoint:
            self.checkpoint = checkpoint
            with self.oplog_progress as oplog_prog:
                oplog_dict = oplog_prog.get_dict()
                # If we have the repr of our oplog collection
                # in the dictionary, remove it and replace it
                # with our replica set name.
                # This allows an easy upgrade path from mongo-connector 2.3.
                # For an explanation of the format change, see the comment in
                # read_last_checkpoint.
                oplog_dict.pop(str(self.oplog), None)
                oplog_dict[self.replset_name] = checkpoint
                LOG.debug("OplogThread: oplog checkpoint updated to %s", checkpoint)
        else:
            LOG.debug("OplogThread: no checkpoint to update.")

    def read_last_checkpoint(self):
        """Read the last checkpoint from the oplog progress dictionary.
        """
        # In versions of mongo-connector 2.3 and before,
        # we used the repr of the
        # oplog collection as keys in the oplog_progress dictionary.
        # In versions thereafter, we use the replica set name. For backwards
        # compatibility, we check for both.
        oplog_str = str(self.oplog)

        ret_val = None
        with self.oplog_progress as oplog_prog:
            oplog_dict = oplog_prog.get_dict()
            try:
                # New format.
                ret_val = oplog_dict[self.replset_name]
            except KeyError:
                try:
                    # Old format.
                    ret_val = oplog_dict[oplog_str]
                except KeyError:
                    pass

        LOG.debug("OplogThread: reading last checkpoint as %s " % str(ret_val))
        self.checkpoint = ret_val
        return ret_val

    def rollback(self):
        """Rollback target system to consistent state.

        The strategy is to find the latest timestamp in the target system and
        the largest timestamp in the oplog less than the latest target system
        timestamp. This defines the rollback window and we just roll these
        back until the oplog and target system are in consistent states.
        """
        # Find the most recently inserted document in each target system
        LOG.debug(
            "OplogThread: Initiating rollback sequence to bring "
            "system into a consistent state."
        )
        last_docs = []
        for dm in self.doc_managers:
            dm.commit()
            last_docs.append(dm.get_last_doc())

        # Of these documents, which is the most recent?
        last_inserted_doc = max(
            last_docs, key=lambda x: x["_ts"] if x else float("-inf")
        )

        # Nothing has been replicated. No need to rollback target systems
        if last_inserted_doc is None:
            return None

        # Find the oplog entry that touched the most recent document.
        # We'll use this to figure where to pick up the oplog later.
        target_ts = util.long_to_bson_ts(last_inserted_doc["_ts"])
        last_oplog_entry = util.retry_until_ok(
            self.oplog.find_one,
            {"ts": {"$lte": target_ts}, "op": {"$ne": "n"}},
            sort=[("$natural", pymongo.DESCENDING)],
        )

        LOG.debug("OplogThread: last oplog entry is %s" % str(last_oplog_entry))

        # The oplog entry for the most recent document doesn't exist anymore.
        # If we've fallen behind in the oplog, this will be caught later
        if last_oplog_entry is None:
            return None

        # rollback_cutoff_ts happened *before* the rollback
        rollback_cutoff_ts = last_oplog_entry["ts"]
        start_ts = util.bson_ts_to_long(rollback_cutoff_ts)
        # timestamp of the most recent document on any target system
        end_ts = last_inserted_doc["_ts"]

        for dm in self.doc_managers:
            rollback_set = {}  # this is a dictionary of ns:list of docs

            # group potentially conflicted documents by namespace
            for doc in dm.search(start_ts, end_ts):
                if doc["ns"] in rollback_set:
                    rollback_set[doc["ns"]].append(doc)
                else:
                    rollback_set[doc["ns"]] = [doc]

            # retrieve these documents from MongoDB, either updating
            # or removing them in each target system
            for namespace, doc_list in rollback_set.items():
                # Get the original namespace
                original_namespace = self.namespace_config.unmap_namespace(namespace)
                if not original_namespace:
                    original_namespace = namespace

                database, coll = original_namespace.split(".", 1)
                obj_id = bson.objectid.ObjectId
                bson_obj_id_list = [obj_id(doc["_id"]) for doc in doc_list]

                # Use connection to whole cluster if in sharded environment.
                client = self.mongos_client or self.primary_client
                to_update = util.retry_until_ok(
                    client[database][coll].find,
                    {"_id": {"$in": bson_obj_id_list}},
                    projection=self.namespace_config.projection(original_namespace),
                )
                # Doc list are docs in target system, to_update are
                # Docs in mongo
                doc_hash = {}  # Hash by _id
                for doc in doc_list:
                    doc_hash[bson.objectid.ObjectId(doc["_id"])] = doc

                to_index = []

                def collect_existing_docs():
                    for doc in to_update:
                        if doc["_id"] in doc_hash:
                            del doc_hash[doc["_id"]]
                            to_index.append(doc)

                retry_until_ok(collect_existing_docs)

                # Delete the inconsistent documents
                LOG.debug("OplogThread: Rollback, removing inconsistent " "docs.")
                remov_inc = 0
                for document_id in doc_hash:
                    try:
                        dm.remove(
                            document_id,
                            namespace,
                            util.bson_ts_to_long(rollback_cutoff_ts),
                        )
                        remov_inc += 1
                        LOG.debug("OplogThread: Rollback, removed %r " % doc)
                    except errors.OperationFailed:
                        LOG.warning(
                            "Could not delete document during rollback: %r "
                            "This can happen if this document was already "
                            "removed by another rollback happening at the "
                            "same time." % doc
                        )

                LOG.debug("OplogThread: Rollback, removed %d docs." % remov_inc)

                # Insert the ones from mongo
                LOG.debug("OplogThread: Rollback, inserting documents " "from mongo.")
                insert_inc = 0
                fail_insert_inc = 0
                for doc in to_index:
                    try:
                        insert_inc += 1
                        dm.upsert(
                            doc, namespace, util.bson_ts_to_long(rollback_cutoff_ts)
                        )
                    except errors.OperationFailed:
                        fail_insert_inc += 1
                        LOG.exception(
                            "OplogThread: Rollback, Unable to " "insert %r" % doc
                        )

        LOG.debug(
            "OplogThread: Rollback, Successfully inserted %d "
            " documents and failed to insert %d"
            " documents.  Returning a rollback cutoff time of %s "
            % (insert_inc, fail_insert_inc, str(rollback_cutoff_ts))
        )

        return rollback_cutoff_ts
